unit UGoal;

// A "TGoal" is a mechanism to compare 2 states and get a "Score" to be
// added to the TPlayer.
//
// A goal can score each little step of the future so the future's score will
// be a sum. of easch step's goal's score, or score can come from comparing
// the initial state of the player with the end state of the future.
//
// Scoring can be a number to sum or take or a reduction coeficient, or both.

interface

uses UState, ExtCtrls;

// Two kind of goals (when used on teams):
//   Individualist:
//     The player that get the goal, get the points.
//     Getting energy from environment is an individualist team goal.
//   Cooperative:
//      All the players in the team get the score.
//      A football team work in cooperative mode.
//   Competitive:
//      The player get the score, the other players get -score.
//      The player that collects more coins wins is competitive.
type TGoalType = ( gtCompetitive = 0, gtIndividualist = 1, gtCooperative = 2);

// Some goals add its scoring at each step, others at the end of the future:
type TGoalFire = (gfStep, gfEnd);


type TGoal = class(TObject)
  private
    // Witch kind of goal are you (only for teams):
    fGoalType: TGoalType;
    // How strong the goal is?
    fStrength: double;
    // Maximum strength limit (defaults to 1):
    fMaxStrength: double;
    // How much (positive or negative) score did I collect while imagining a future?
    fScore: double;
    // Reductive goals reduce score by using a coef. 0-1 like Score x 0.9
    fReduce: double;
    // When adding score to a player, I keep it on a variable:
    fPlayer: TObject;
    // You get more points if you get them soon in the future:
    // Ops! Bad idea, scoring must be done at the future's end, not at each step
    // but I keep this experimental feature, you never knows.
    fSoonerIsBetter: boolean;
    fAlwaysDraw: boolean;
    procedure AddScore;
    procedure SetStrength(v: double);
  protected
    // Must this goal be drawed for each player (or only once for all players)?
    fDrawable: boolean;
    // Must it be drawed when imagining or when in "real life"?
    fDrawOnImagining: boolean;
    // To show it to the user I need a name:
    fDisplayName: string;
    procedure CalcScore(FireAt: TGoalFire); virtual; abstract;
    procedure Draw; virtual;
    property  Player: TObject read fPlayer write fPlayer;
  public
    constructor create(c: double); virtual;
    procedure ScoreGoal(v: TObject; FireAt: TGoalFire); virtual;
    // Goals with a "state", like drops, needs a second "state" while imagining.
    procedure ResetImaginaryState; virtual;
    // When a goal doesn't draw with each player (like drops) you call it to draw:
    procedure NeutralDraw; virtual;
    property  AlwaysDraw: boolean read fAlwaysDraw write fAlwaysDraw;
    property  Strength: double read fStrength write SetStrength;
    property  MaxStrength: double read fMaxStrength write fMaxStrength;
    property  GoalType: TGoalType read fGoalType write fGoalType;
    property  Score: double read fScore write fScore;
    property  Reduce: double read fReduce write fReduce;
    property  Drawable: boolean read fDrawable;
    property  DrawOnImagining: boolean read fDrawOnImagining;
    property  DisplayName: string read fDisplayName;
    property  SoonerIsBetter: boolean read fSoonerIsBetter write fSoonerIsBetter;
  end;


implementation

uses Math, UPlayer, UTrack;


constructor TGoal.create(c: double);
begin
  inherited create;
  fPlayer:= Nil;
  fGoalType:= gtIndividualist;
  fScore:= 0;
  fReduce:= 1;
  fStrength:=  c;
  fMaxStrength:= 10.0; // default max value for strength.
  fSoonerIsBetter:= false;
  // By default, try to draw on real move and once (not for each player):
  fDrawable:= false;
  fDrawOnImagining:= false;
  fDisplayName:= 'Generic goal';
  fAlwaysDraw:= false;
end;

procedure TGoal.ResetImaginaryState;
begin
  // Goals that have an state, like drops, need to have a secondary imaginary state
  // and reset it every time a new future starts. Only if player is not imaginig
  // the real state will change.
end;

procedure TGoal.SetStrength(v: double);
begin
  fStrength:= Max(0.0, Min(v, MaxStrength));
end;

procedure TGoal.ScoreGoal(v: TObject; FireAt: TGoalFire);
var
  TimeCoef: double;
begin
  if (Strength > 0) then begin
    Score:= 0;
    Reduce:= 1;
    Player:= v;
    CalcScore(FireAt);
    // Some step goals want to get the score as sooner as possible (deprecated):
    if (FireAt=gfStep) and SoonerIsBetter and TPlayer(Player).Imagining then begin
      // Build a "TimeCoef" to make inminent things score stronger:
      //   Time passed =   0% of total time to simulate --> TimeCoef = 1.0^2 = 1.000
      //   Time passed =  20% of total time to simulate --> TimeCoef = 0.8^2 = 0.064
      //   Time passed =  50% of total time to simulate --> TimeCoef = 0.5^2 = 0.025
      //   Time passed =  80% of total time to simulate --> TimeCoef = 0.2^2 = 0.004
      //   Time passed = 100% of total time to simulate --> TimeCoef = 0.0^2 = 0.000
      TimeCoef:= Sqr(1- TPlayer(Player).State.TimePassed / TTrack(TPlayer(Player).Track).FutureMSeconds);
      // Reduce scoring as time passes:
      if (Score<>0) then
        Score:= Score * TimeCoef;
      // If it is a reduction goal:
      if (Reduce < 1) then
        Reduce:= Reduce + (1-Reduce) * TimeCoef;
    end;
    if Drawable and TPlayer(player).DebugGoals then
      Draw;
    AddScore;
  end;
end;

procedure TGoal.AddScore;
begin
  // Add scoring to the player:
  TPlayer(Player).State.AddScore(Score * Strength, Power(Reduce, 1+Strength));
  // Reset internal scoring:
  Score:= 0;
  Reduce:= 1;
end;

// Draw the player's goal scoring:
procedure TGoal.Draw;
begin
  // By default a goal draws nothing.
end;

// Draw the general (no player needed) picture (for instance, drops)
procedure TGoal.NeutralDraw;
begin
  // Goals with Personalized Drawing need to redefine this.
end;

end.
