unit UKart;

interface

uses
  ExtCtrls, Graphics, Math, Types, DateUtils, SysUtils, Contnrs, Forms,
  UTrack2D, UState, UPlayer, UPlayer2D, UUtils;

type
  TKart = class(TPlayer2D)
  private
    // State params' indexes:
    fTurn, fPower, fPushTurn, fPushPower: integer;
    // State components set/get:
    procedure SetTurn(v: double);
    function  GetTurn: double;
    procedure SetPower(v: double);
    function  GetPower: double;
    function  GetPushTurn: double;
    function  GetPushPower: double;
    // Limits of state components:
    procedure SetMaxPower(v: double);
    function  GetMaxPower: double;
    procedure SetMinPower(v: double);
    function  GetMinPower: double;
    procedure SetMaxTurn(v: double);
    function  GetMaxTurn: double;
  protected
    // Define your state components:
    procedure DefineStateParams; override;
    // Update the points on screen that form the player's siluette:
    procedure CalcNodes; override;
    // I only want left/right futures to be shown:
    function  DebugThisFuture: boolean; override;
    function  MakeADuplicate: TPLayer; override;
  public
    constructor create(x, y, a: double);
    procedure Simulate; override;
    procedure draw; override;
    function  DebugColor(ThisFuture: TState): TColor; override;
    // State params:
    property Turn:      double  read GetTurn      write SetTurn;
    property Power:     double  read GetPower     write SetPower;
    property PushTurn:  double  read GetPushTurn;
    property PushPower: double  read GetPushPower;
    // Limits of computed params:
    property MaxPower:  double  read GetMaxPower  write SetMaxPower;
    property MinPower:  double  read GetMinPower  write SetMinPower;
    property MaxTurn:   double  read GetMaxTurn   write SetMaxTurn;
  end;


implementation

uses UTrack;


// ****************************
// **  ADMINISTRATIVE STUFF  **
// ****************************

constructor TKart.create(x, y, a: double);
begin
  inherited create;
  PlaceHere(x, y, a);
  Size:= 10;
  fTargetFutures:= 80;
  SetIniState;
end;

procedure TKart.DefineStateParams;
begin
  // PosX and PosY are mandatory, so they are defined on TPLayer2D:
  inherited;
  // Add state params: (Name, Type, Value, Max, Min, Precision)
  State.AddParam( TStateParam.createComputed('Turn',  0, PI/4, -PI/4) );
  State.AddParam( TStateParam.createComputed('Power', 0,   15,  -15) );
  State.AddParam( TStateParam.createFree('PushTurn',   15) );
  State.AddParam( TStateParam.createFree('PushPower', 180) );
  // State param's Indexes:
  fTurn:=      State.ParamIndexByName('Turn');
  fPower:=     State.ParamIndexByName('Power');
  fPushTurn:=  State.ParamIndexByName('PushTurn');
  fPushPower:= State.ParamIndexByName('PushPower');
end;


// *********************
// **  DRAWING STUFF  **
// *********************

procedure TKart.CalcNodes;
var i: integer;
const
  Angles: array[0..3] of double = (-0.1, 0.1, 0.8, -0.8);
begin
  // Nodes is a dynamic array, so I must set size on first use.
  if NodesNeedIni then begin
    SetLength(Nodes, 4);
    SetLength(NodesR, 4);
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= Low(Nodes) to High(Nodes) do begin
      NodesR[i, 1]:= PosX + Size*cos(Angle + Angles[i]*PI);
      NodesR[i, 2]:= PosY + Size*sin(Angle + Angles[i]*PI);
      Nodes[i].X:= round(NodesR[i,1]);
      Nodes[i].Y:= round(NodesR[i,2]);
    end;
    NodesNeedRecalc:= false;
    VectorsNeedRecalc:= true;
  end;
end;

procedure TKart.draw;
var
  i: integer;
  Diam: integer;
  mx, my, Joy_X, Joy_Y: double;
  TyreIni, TyreEnd: TPoint;
  Track2D: TTrack2D;
begin
  // TPlayer will draw the body for me using the nodes:
  inherited;
  Track2D:= TTrack2D(Track);
  // TIRES
  // =====
  // Add the 2 front tires only (so you can see where is it heading to):
  for i:= 0 to 1 do begin
    // Tyres are just lines:
    TyreIni.X:= round(PosX + Size*0.8*(cos(Angle + (i*2-1) * 0.3*PI)));
    TyreIni.Y:= round(PosY + Size*0.8*(sin(Angle + (i*2-1) * 0.3*PI)));
    TyreEnd.X:= TyreIni.X + round(Size*0.6*(cos(Angle + Turn)));
    TyreEnd.Y:= TyreIni.Y + round(Size*0.6*(sin(Angle + Turn)));
    // Draw them:
    Track2D.Img.Canvas.Pen.Color:= clBlack;
    Track2D.Img.Canvas.Pen.Width:= 2;
    Track2D.Img.Canvas.MoveTo(TyreIni.X, TyreIni.Y);
    Track2D.Img.Canvas.LineTo(TyreEnd.X, TyreEnd.Y);
    Track2D.Img.Canvas.Pen.Width:= 1;
  end;
  // JOYSTICK
  // ========
  Track2D.Img.Canvas.Brush.Color:= clWhite;
  // The position change with power and turn:
  if (Power>=0) then
    mx:= Power/MaxPower
  else
    mx:= Power/Abs(MinPower);
  my:= Turn/MaxTurn;
  Joy_X:= PosX + (Cos(Angle)*mx*0.6 - Sin(Angle)*my) * Size;
  Joy_Y:= PosY + (Sin(Angle)*mx*0.6 + Cos(Angle)*my) * Size;
  Diam:= round(Size * 0.4);
  // Draw the stick:
  Track2D.Img.Canvas.Pen.Width:= 2;
  Track2D.Img.Canvas.MoveTo(round(PosX), round(PosY));
  Track2D.Img.Canvas.LineTo(round(Joy_X), round(Joy_Y));
  Track2D.Img.Canvas.Pen.Width:= 1;
  // Draw the stick's ball: Fill it red (power<0) or green (power>0):
  if (Power>=0) then begin
    //Green to black
    i:= round(255 * Sqrt(Power/MaxPower));
    Track2D.Img.Canvas.Brush.Color:= RGBToColor(255-i, 255, 255-i);
  end else begin
    //Yellow to black
    i:= round(255 * Sqrt(Power/MinPower));
    Track2D.Img.Canvas.Brush.Color:= RGBToColor(255, 255, 255-i);
  end;
  Track2D.Img.Canvas.Pen.Color:=   clLtGray;
  Track2D.Img.Canvas.Ellipse(round(Joy_X-Diam), round(Joy_Y-Diam), round(Joy_X+Diam), round(Joy_Y+Diam));
  Track2D.Img.Canvas.Pen.Color:=   clWhite;
  Track2D.Img.Canvas.Brush.Color:= clWhite;
  Track2D.Img.Canvas.Ellipse(round(Joy_X-Diam*0.8), round(Joy_Y-Diam*0.8),
                             round(Joy_X+Diam*0.2), round(Joy_Y+Diam*0.2));
  Track2D.Img.Canvas.Brush.Style:= bsClear;
  Track2D.Img.Canvas.Pen.Color:=   clBlack;
  Track2D.Img.Canvas.Ellipse(round(Joy_X-Diam), round(Joy_Y-Diam), round(Joy_X+Diam), round(Joy_Y+Diam));
end;

// Debug only left/right decision, not run/stop ones:
function TKart.DebugThisFuture: boolean;
begin
  result:= (State.Param[fPushTurn].IniValue <> 0);
end;

// Debug lines are blue/red for left/right decisions:
function TKart.DebugColor(ThisFuture: TState): TColor;
var C1, C2: integer;
begin
  // Red or blue color for left/right push:
  C1:= 255;
  // As future goes away it fades to white:
  C2:= Round( 155 + 100 * ThisFuture.TimePassed / TTrack(Track).FutureMSeconds);
  if (ThisFuture.Param[fPushTurn].IniValue > 0) then
    // Right turning is red:
    result:= RGBToColor(C1, C2, C2)
  else
    // Left turning is blue:
    result:= RGBToColor(C2, C2, C1);
end;


// ***********************
// **  SIMULATION STUFF **
// ***********************

// Let run the simulation for a delta time and update state acordingly:
procedure TKart.Simulate;
var
  // Time pased
  s, s2: double;
  // Acceleration components:
  Acc_X, Acc_Y, UsePower, UseEnergy: double;
  // Need vectors to decompose velocity:
  Va_L, Va_X, Va_Y, Vn_L, Vn_X, Vn_Y: double;
  // Friction coefs. for both velocities:
  UseFriction, UseSideFriction: double;
  Friction: double;
const
  MinEnergy: double = 0.05; // Stoped it spend this energy per second.
begin
  // Inherited will decide if I am dead, out of track, etc.:
  inherited;
  // Milliseconds to seconds:
  s:= SimSeconds;
  s2:= Sqr(s)/2;
  // APPLY DECISION
  // ==============
  // Dead karts dont push power!
  if dead then begin
    Power:= 0;
  end else begin
    Turn:=  Turn  + PushTurn  * s;
    Power:= Power + PushPower * s;
  end;
  // VELOCITY
  // ========
  // Decompose vel. into "ahead vel" (Va) and normal vel (Vn):
  // Ahead vel.: Length is first calc, then get its components:
  Va_L:= VelX * Cos(Angle) + VelY * Sin(Angle);
  Va_X:= Va_L * Cos(Angle);
  Va_Y:= Va_L * Sin(Angle);
  // Now the "rest" of vel. will be "normal velocity":
  Vn_X:= VelX - Va_X;
  Vn_Y:= VelY - Va_Y;
  // FRICTION COEFS
  // ==============
  UseFriction:=     TTrack2D(Track).Friction;
  UseSideFriction:= TTrack2D(Track).SideFriction;
  // Convert to a 0-1 coeficient:
  Friction:=  (1 + 0.01*Va_L) *  UseFriction;
  Friction:=  Max(0, (1- (s*10 * Friction )));
  // POWER FOR ENGINE AND ENERGY SPENT
  // =================================
  // Usual drive:
  UsePower:=  Power;
  // Energy spent per second: A fixed "living" cost + energy for used power:
  UseEnergy:= MinEnergy + Sqr(Abs(Power)/10)/10;
  if (Power<0) then begin
    // Power negative and velocity positive: you are breaking:
    if (Va_L > 0) then begin
      // Braking takes out a fixed amount to the velocity length:
      // At full braking, I take off 100 times the friction on the velocity.
      Va_L:= Max(0, Va_L - s*UseFriction* 100 *MaxPower*(Abs(Power/MinPower)));
      Va_X:= Va_L * Cos(Angle);
      Va_Y:= Va_L * Sin(Angle);
      UsePower:=  0;
      UseEnergy:= MinEnergy;
    end else begin
      // Driving backwards? OK, but energy eficiency is 20% only:
      UsePower:= Power/10;
    end;
  end;
  // Used energy = seconds * energy per second
  UseEnergy:= UseEnergy * s;

  {
  if not Imagining then begin
    TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
    TTrack2D(Track).Img.Canvas.TextOut(Round(PosX+10), Round(PosY), 'E='+Format('%3.3f', [UseEnergy]));
    TTrack2D(Track).Img.Canvas.TextOut(Round(PosX+10), Round(PosY+10), 'P='+Format('%3.3f', [Abs(Power)/10]));
  end;
  }

  // Can't use more energy than available:
  if (UseEnergy > Energy) then begin
    UsePower:= UsePower * (Energy/UseEnergy);
    UseEnergy:= Energy;
  end;
  // Power depletes your energy:
  UsePower:= Sign(UsePower) * Min(Energy*200, Abs(UsePower));
  Energy:= Energy - UseEnergy;
  // No energy, no living:
  if Energy=0 then Dead:= true;
  // SLIDING CONTROL
  // ===============
  // Is like braking but with a fixed "power" and a maximum g.
  Vn_L:= Sqrt(Sqr(Vn_X) + Sqr(Vn_Y));
  if (Vn_L>0.0001) then begin
    // Normalize:
    Vn_X:= Vn_X / Vn_L;
    Vn_Y:= Vn_Y / Vn_L;
    // Adjust length:
    Vn_L:= Max(0, Vn_L - 100*s*UseSideFriction);
    // Recalc componets:
    Vn_X:= Vn_X * Vn_L;
    Vn_Y:= Vn_Y * Vn_L;
  end;
  // UPDATES
  // =======
  // Angle:
  Angle:= Angle + Turn * s * Min(1, Va_L/10);
  // Acceleration (forward or backward):
  Acc_X:= UsePower * Cos(Angle);
  Acc_Y:= UsePower * Sin(Angle);
  // Velocity:
  VelX:= (Va_X*Friction + Vn_X) + s * Acc_X;
  VelY:= (Va_Y*Friction + Vn_Y) + s * Acc_Y;
  // Position:
  // Remember: Pos(t+s) = Pos(t) + Vel(t)*s + Acc(t)*s^2
  PosX:= PosX + VelX * s + Acc_X * s2;
  PosY:= PosY + VelY * s + Acc_Y * s2;
  // Update 4 nodes of the kart skeleton:
  NodesNeedRecalc:= true;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

function TKart.MakeADuplicate: TPlayer;
begin
  // Duplicate yourself:
  result:= TKart.create(0, 0, 0);
  TKart(result).CopyFrom(self);
end;

// *******************
// ** SET/GET STUFF **
// *******************

procedure TKart.SetTurn(v: double);
begin
  State.Param[fTurn].Value:= v;
end;
function  TKart.GetTurn: double;
begin
  result:= State.Param[fTurn].Value;
end;

procedure TKart.SetPower(v: double);
begin
  State.Param[fPower].Value:= v;
end;
function  TKart.GetPower: double;
begin
  result:= State.Param[fPower].Value;
end;

procedure TKart.SetMaxPower(v: double);
begin
  State.Param[fPower].MaxValue:= v;
end;
function  TKart.GetMaxPower: double;
begin
  result:= State.Param[fPower].MaxValue;
end;

procedure TKart.SetMinPower(v: double);
begin
  State.Param[fPower].MinValue:= v;
end;
function  TKart.GetMinPower: double;
begin
  result:= State.Param[fPower].MinValue;
end;

procedure TKart.SetMaxTurn(v: double);
begin
  State.Param[fTurn].MaxValue:= v;
end;
function  TKart.GetMaxTurn: double;
begin
  result:= State.Param[fTurn].MaxValue;
end;

function  TKart.GetPushTurn: double;
begin
  result:= State.Param[fPushTurn].Value;
end;

function  TKart.GetPushPower: double;
begin
  result:= State.Param[fPushPower].Value;
end;


end.
