unit UListOfFutures;

// Intro:
// ======
// When a player "imagines" a new possible future, it finally get a "State" that
// contains both the final state of the player (on its computed StrateParam's
// Values) plus the initial decision taken in order to later by able to group
// futures based on this initial decision (free params' ini values).
//
// All this list of futures and related stuff is defined here.
//
// The main reason to have this unit is to keep all the heart of the intelligence
// out of the other stuff:
//
// 1) Detect "options", the list of initial decisions used on the futures.
// 2) Group futures into its corresponding "option".
// 3) Discard futures with same ending point (MacroState) on each option.
// 4) Score options by taking the higest score among its futures.
// 5) Take the intelligent decision based on options and their scorings.

interface

uses Contnrs, Graphics,
     UState;


type TPointList = array of array of double;

type TListOfFutures = class(TObject)
  private
    fOwner: TObject; // is a TPlayer
    // The list of futures states:
    fFutures: TObjectList;
    // We will ned the number of different futures found:
    fDifferentFutures: integer;
    // List of "options" = different initial decisions in the futures
    fOptions: TObjectList;
    function GetFuture(i: integer): TState;
    function GetOption(i: integer): TState;
    function GetOptionCount: integer;
    function GetFutureCount: integer;
  protected
  public
    constructor create(v: TObject);
    destructor destroy; override;
    // Add or take futures:
    procedure AddFuture(v: TState);
    procedure Clear;
    // Debugging:
    procedure DrawOptions;
    // Thinking proccess as a whole:
    procedure TakeIntelligentDecision(DestinationState: TState);
    // Now the parts of the proccess:
    procedure ExtractOptions;
    procedure ScoreOptions;
    procedure NormalizeOptions;
    procedure AverageOptions(DestinationState: TState);
    procedure RefineDecision(DestinationState: TState);
    procedure AutoAdjustSensitivity(DestinationState: TState);
    // Math utils for getting a spline max. values
    procedure SplineMaxValue(Points: TPointList; var MaxX, MaxY, AvgScore, AvgValue: double);
    property  Owner: TObject read FOwner;
    // Property to count futures/options:
    property  OptionCount: integer read GetOptionCount;
    property  FutureCount: integer read GetFutureCount;
    property  DifferentFutures: integer read fDifferentFutures write fDifferentFutures;
    // Properties to access a given future/option as a TState:
    property  Future[i: integer]: TState read GetFuture;
    property  Option[i: integer]: TState read GetOption;
  end;


implementation

uses Classes, Math, Forms, SysUtils,
     UPlayer, UTrack;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TListOfFutures.create(v: TObject);
begin
  inherited create;
  fOwner:= v;
  DifferentFutures:= 0;
  fFutures:= TObjectList.create;
  fFutures.OwnsObjects:= false;
  fOptions:= TObjectList.create;
  fOptions.OwnsObjects:= false;
end;

destructor TListOfFutures.destroy;
begin
  // Clear futures:
  while fFutures.Count > 0 do begin
    TState(fFutures.Items[0]).Free;
    fFutures.Delete(0);
  end;
  fFutures.free;
  // Clear options:
  while fOptions.Count > 0 do begin
    TStateParam(fOptions.Items[0]).Free;
    fOptions.Delete(0);
  end;
  fOptions.free;
  inherited;
end;

procedure TListOfFutures.AddFuture(v: TState);
var
  NewFuture: TState;
begin
  NewFuture:= TState.create(v.owner);
  NewFuture.CopyFrom(v);
  // Convert micro state into a macro state (round positional params):
  NewFuture.MicroToMacro;
  fFutures.Add(NewFuture);
end;

procedure TListOfFutures.Clear;
begin
  while fFutures.Count > 0 do begin
    TState(fFutures.Items[0]).Free;
    fFutures.Delete(0);
  end;
  while fOptions.Count > 0 do begin
    TStateParam(fOptions.Items[0]).Free;
    fOptions.Delete(0);
  end;
end;

function TListOfFutures.GetFuture(i: integer): TState;
begin
  result:= TState(fFutures[i]);
end;

function TListOfFutures.GetOption(i: integer): TState;
begin
  result:= TState(fOptions[i]);
end;

function TListOfFutures.GetOptionCount: integer;
begin
  result:= fOptions.Count;
end;

function TListOfFutures.GetFutureCount: integer;
begin
  result:= fFutures.Count;
end;


// ********************
// ** THINKING STUFF **
// ********************

// OK, I have a list of futures and they are scored, I need to group them
// by comparing their initial decisions, the discard repeated ones, then
// sum scores by group, normalize those scores, and average initial decision
// of each group by its score. This average is "the intelligent" behaviour
// as defined by "causal enthropic forces".
//
// It is important to note that the "Team" this player plays in can act as
// a re-scoring machine so my decision will be biased by my team scoring,
// but all this, if happends, does before this point, so I can freely delete
// or reorder my futures without affecting outer proccess.

// Decision taking proccess:
procedure TListOfFutures.TakeIntelligentDecision(DestinationState: TState);
begin
  // Make a list of all different initial decisions or "options":
  ExtractOptions;
  // Score each option with the different futures belonging to it:
  ScoreOptions;
  // Normalize the option scores so they sum 1:
  NormalizeOptions;
  // Average option*score will produce the standard "intelligent decisions":
  AverageOptions(DestinationState);
  // Refine decision to avoid indecision in case of a "Y shaped" bifurcation.
  // Actualy deprecated, not so good idea.
  if TPlayer(fOwner).UseRefinedAI then
    RefineDecision(DestinationState);
  // AutoAdjustSensitivity: if both small options both are gretater than extreme
  // values, then the sweet point -option with maximum entropy or score- is near
  // the center and zooming in is neccesary to gain fine control on smooth like
  // situation like takiing land safely:
  if TPlayer(fOwner).AutoAdjustSensitivity then
    AutoAdjustSensitivity(DestinationState);
end;

// For each different initial decision I create a "Option", flag all the futures
// starting with this option as belonging to it:
procedure TListOfFutures.ExtractOptions;
var
  f, o, MyOption: integer;
  NewOption: TState;
begin
  fOptions.Clear;
  // Assign each future to an "option" depending on the initial decision:
  for f:= 0 to FutureCount-1 do begin
    // Is this initial decision in my option list?
    MyOption:= -1;
    for o:= 0 to fOptions.Count-1 do begin
      if Future[f].HaveSameInitialDecision(Option[o]) then begin
        MyOption:= o;
        break;
      end;
    end;
    // If not found, add it:
    if (MyOption = -1) then begin
      NewOption:= TState.create(owner);
      NewOption.CopyFrom(Future[f]);
      NewOption.ResetScore;
      fOptions.Add(NewOption);
      MyOption:= fOptions.Count-1;
    end;
    // Connect this future with the option it took at first decision:
    Future[f].BelongsTo:= Option[MyOption];
    // Emergency debugging:
    //TPlayer(fOwner).DrawFutureScore(Future[f]);
  end;
  // Emergency debugging:
  //Application.ProcessMessages;
  //TTrack(TPlayer(fOwner).Track).Draw(true);
end;

procedure TListOfFutures.ScoreOptions;
var
  f, f2, o: integer;
  AlreadyFound: boolean;
  Opt: TState;
begin
  DifferentFutures:= 0;
  // Set scoring to zero for all options:
  for o:= 0 to fOptions.Count-1 do
    TState(Option[o]).ResetScore;
  // For each future found:
  for f:= 0 to FutureCount-1 do begin
    // Witch option correspond to this future?
    Opt:= TState(TState(Future[f]).BelongsTo);
    // The final position of this future is new for the option it belongs to or
    // was already found ?
    AlreadyFound:= false;
    for f2:= 0 to (f-1) do begin
      // Same start and same end = same future, so if it not discarted, we have
      // two paths for the same different future... witch one should I keep?
      if (not Future[f2].FutureDiscarted) and
         (TState(Future[f2]).BelongsTo = Opt) and
         (Future[f].HaveSameFinalState(Future[f2])) then begin
        AlreadyFound:= true;
        // Witch future is better, f or f2? ...the one with the higer scoring.
        if (TState(Future[f]).Score > TState(Future[f2]).Score) then begin
          // This f is better than f2 to represent this final point:
          Future[f2].FutureDiscarted:= true; // f2 is downgraded to "discarted future"
          // Update option's score with the difference:
          Opt.AddScore( TState(Future[f]).Score - TState(Future[f2]).Score );
        end else begin
          // Mark this future f as discarted, it is better represented by f2:
          Future[f].FutureDiscarted:= true;
        end;
        break;
      end;
    end;
    // If it is a brand new future, add it to the list:
    if not AlreadyFound then begin
      DifferentFutures:= DifferentFutures + 1;
      // Found a new different future for this option:
      Opt.AddScore(Future[f].Score);
      // Number of different futures is used as score on AI levels 1 to 3:
      Opt.HitCount:= Opt.HitCount + 1;
    end;
    // Emergency debugging:
    //TPlayer(fOwner).DrawFutureScore(Future[f]);
  end;
  // Emergency debugging:
  //Application.ProcessMessages;
  //TTrack(TPlayer(fOwner).Track).Draw(true);
end;

// Divide each option.score by sum(option.score) so they sum 1:
// Note: if options score zero, you have no goals and Log(N) (AI Level 3) is used
procedure TListOfFutures.NormalizeOptions;
var
  o: integer;
  SumOfScores, MinScore: double;
begin
  SumOfScores:= 0;
  // AI Level 7: Adjust options' scores so lowest is zero:
  if TPlayer(fOwner).UseZeroedAI then begin
    // Get the min. of all options found:
    MinScore:= Option[0].Score;
    for o:= 1 to fOptions.Count-1 do
      MinScore:= Min(MinScore, Option[o].Score);
  end;
  // Get the sum of all options found:
  for o:= 0 to fOptions.Count-1 do
    SumOfScores:= SumOfScores + Option[o].Score;
  // Especial case: You didnt choose Level3 AI, but there is no goal so scores
  // are all zero, in this case, only Lon(N) from AI Level 3 can be used:
  if SumOfScores=0 then begin
    SumOfScores:= 0;
    for o:= 0 to fOptions.Count-1 do begin
      Option[o].SetScore( Ln(Option[o].HitCount) );
      SumOfScores:= SumOfScores + Option[o].Score;
    end;
  end;
  // Divide option's score by its sum so they now sum 1 (normalize):
  if (SumOfScores > 0) then
    for o:= 0 to fOptions.Count-1 do
      Option[o].SetScore( Option[o].Score / SumOfScores );
end;

// Sum(option * score) will form the "intelligent" decision:
// Note: As I added a new way to decide to refine this, I will call this decision
// the "averaged" decision.
procedure TListOfFutures.AverageOptions(DestinationState: TState);
var p, o: integer;
begin
  // Now I have "relevant" futures (options) one per initial decison and a
  // different future end (macrostate), so my final decision will be:
  //   Sum[Initial_decision * score^2 / Sum(score^2) ]
  // As scores have been squared and then normalized, it simplifies to:
  //   Sum[Initial_decision * score]
  //
  // It is calculated for each free param, and finally you get a list of values
  // for each free param, those values are stored in a DestinationState, that
  // usually means in TPlayer.ActualState, so the TPlayer.Simulate can retrieve
  // this decision and apply it to change TPlayer.ActualState as time passes.
  //
  // For each free param on the state, set it to zero and average option's values
  // weightened by option's score:
  for p:= 0 to DestinationState.FreeParamCount-1 do begin
    // OK, set to zero before start accumulating:
    DestinationState.FreeParam[p].Value:= 0.0;
    // Accumulate on all options found:
    for o:= 0 to fOptions.Count-1 do begin
      DestinationState.FreeParam[p].Value:= DestinationState.FreeParam[p].Value +
          Option[o].Score * Option[o].FreeParam[p].IniValue;
    end;
  end;
end;

// AI Level 6 (deprecated):
// ========================
// You have the weightened average as the free param decision as paper suggest,
// now we refine it avoiding "Y shaped" indecisions: we construct the spline
// passing by the points (option.freeparamvalue, option.score) for all options
// with freeparam value not zero. From the spline, get the Y "score" for the
// averaged decision and also the max. score on the spline, and mix the two.
//
// One mixing is to decide 0.7*averaging + 0.3*maximized, but a better aproach
// is to give more weigth to the maximized decision as its score gets bigger
// compared to averaged scored, so if max. gives 300 score and averaged decision
// get 100, then use:
//   Decision = Maximized * (300/(300+100)) + averaged * (100/(300+100))
procedure TListOfFutures.RefineDecision(DestinationState: TState);
var
  p, o, i, j, k: integer;
  Tmp: double;
  // Points will stored on 2 dimensional double arrays:
  Points: TPointList;
  PointCount: integer;
  // Maximized decision:
  MaxX, MaxY: double;
  // Score for the actual "averaged" decision:
  AvgScore, AvgValue: double;
  // Mixing coef. for averaged decision (take the rest from maximized decision):
  AvgCoef: double;
begin
  // For each free param on the state, set it to zero and take a decision.
  for p:= 0 to DestinationState.FreeParamCount-1 do begin
    // Clear the list of points:
    PointCount:= 0;
    SetLength(Points, 0);
    // Add a points for each options found:
    for o:= 0 to FOptions.Count-1 do begin
      // This option has a non-zero value for this free param?
      if (Option[o].FreeParam[p].IniValue <> 0) then begin
        // Add it to the list of points:
        inc(PointCount);
        SetLength(Points, PointCount);
        SetLength(Points[PointCount-1], 2);
        Points[PointCount-1, 0]:= Option[o].FreeParam[p].IniValue;
        Points[PointCount-1, 1]:= Option[o].Score;
      end;
    end;
    // Needed 3 points at least to make sense.
    if (PointCount>2) then begin
      // Reorder points naturally (lower free param values go first):
      for i:= 1 to PointCount-1 do begin
        for j:= (i-1) downto 0 do begin
          if (Points[j,0] > Points[j+1,0]) then begin
            for k:= 0 to 1 do begin
              Tmp:= Points[j,k];
              Points[j,k]:= Points[j+1,k];
              Points[j+1,k]:= Tmp;
            end;
          end;
        end;
      end;
      // Get a spline with those points and give me back the max.Y -score- and
      // its X (free param's value).
      // I also want to know the score for a know value: the averaged decision
      AvgValue:= DestinationState.FreeParam[p].Value;
      SplineMaxValue(Points, MaxX, MaxY, AvgScore, AvgValue);
      // Mix this MaxX value with the Averaged value to get the final decision.
      // Coef. for the mix comes from comparing scores of both decisions (Agv/Max)
      // but I added Max(0.001, value) to both to avoid divisions by zero in
      // extreme cases where player is near to crashing.
      //if MaxY > 3*AvgScore then begin
        AvgCoef:= sqrt(Max(0.001, AvgScore)/Max(0.001, MaxY));
        DestinationState.FreeParam[p].Value:= AvgValue * AvgCoef + MaxX * (1-AvgCoef);
      //end;
    end;
  end;
end;

// AutoAdjustSensitivity: if both small options both are gretater than extreme
// values, then the sweet point -option with maximum entropy or score- is near
// the center and zooming in is neccesary to gain fine control on smooth like
// situation like takiing land safely:
procedure TListOfFutures.AutoAdjustSensitivity(DestinationState: TState);
var
  p, o, n: integer;
  MinSmall, MaxSmall, MaxBig, MinBig, IniValue: double;
begin
  // For each free param on the state.
  for p:= 0 to DestinationState.FreeParamCount-1 do begin
    // MinValue of small options > MaxValue of max options, then zoom in:
    MinSmall:= -1;
    MaxSmall:= 0;
    MinBig:= -1;
    MaxBig:= 0;
    n:= 0;
    // Check max and min scorings for small and big values for this free param
    // by looking at options that match:
    for o:= 0 to fOptions.Count-1 do begin
      if (Option[o].FreeParam[p].IniValue <> 0) then begin
        inc(n);
        // Small or Max ini value?
        IniValue:= Abs(Option[o].FreeParam[p].IniValue);
        if ( Abs(IniValue - Option[o].FreeParam[p].SmallOption) < Abs(IniValue - Option[o].FreeParam[p].BigOption) ) then begin
          // Update small limits:
          if (MinSmall = -1) then begin
            MinSmall:= Option[o].Score;
            MaxSmall:= Option[o].Score;
          end else begin
            MinSmall:= Min(MinSmall, Option[o].Score);
            MaxSmall:= Max(MaxSmall, Option[o].Score);
          end;
        end else begin
          // Update big limits:
          if (MinBig = -1) then begin
            MinBig:= Option[o].Score;
            MaxBig:= Option[o].Score;
          end else begin
            MinBig:= Min(MinBig, Option[o].Score);
            MaxBig:= Max(MaxBig, Option[o].Score);
          end;
        end;
      end;
    end;
    // Found 4 values? 2 valued free params can't be adjusted!
    if (n=4) then begin
      if (MaxSmall > MaxBig) then begin
        // Zooming in (higher sensitivity when chossing options for this free param):
        if (DestinationState.FreeParam[p].Sensitivity > 0.1) then
          DestinationState.FreeParam[p].Sensitivity:= DestinationState.FreeParam[p].Sensitivity * 0.9;
      end else if (MaxBig > MaxSmall) then begin
        // Zoom out (lower sensitivity when chossing options for this free param):
        if (DestinationState.FreeParam[p].Sensitivity < 1) then
          DestinationState.FreeParam[p].Sensitivity:= Min(1, DestinationState.FreeParam[p].Sensitivity * 1.3);
      end;
    end;
  end;
end;

// Using the "Cubic Hermite Spline"
procedure TListOfFutures.SplineMaxValue(Points: TPointList; var MaxX, MaxY, AvgScore, AvgValue: double);
var
  i, j: integer;
  i0, i1, i2, i3: integer;
  c1, c2, c3, c4, x, y, t, t2, t3: double;
  r: array[1..2] of array[1..2] of double;
  OldX: double;
begin
  // Ini the averaged decision's score:
  AvgScore:= -9999;
  // In order to detect when spline pass by the AvgValue, I keep track of last
  // point in spline's x:
  OldX:= Points[0,0];
  // Note: Initial and final points are considered "doubled" so spline formulae
  // is simplier, as for each point i I can use the previous and following points.
  MaxX:= 0.0;
  MaxY:= 0.0; // Y is supposed to be possitive, so zero is a good starting point.
  // For each section of the spline connecting points i and i+1...
  for i:= 0 to High(Points)-1 do begin
    // Analize the section of the spline that goes from i to i+1
    // I need 4 points: i0=previous to i, i1=i, i2=i+1 and i3=following i+1
    i0:= Max(i-1, 0);
    i1:= i;
    i2:= i+1;
    i3:= Min(i+1, High(Points));
    // Vector connecting i1-1 and i1+1
    r[1,1]:= 0.5*(Points[i2,0] - Points[i0,0]);
    r[1,2]:= 0.5*(Points[i2,1] - Points[i0,1]);
    // Vector connecting i2-1 and i2+1
    r[2,1]:= 0.5*(Points[i3,0] - Points[i1,0]);
    r[2,2]:= 0.5*(Points[i3,1] - Points[i1,1]);
    // Calc. 6 points inside this section looking for the max. Y value:
    for j:=0 to 5 do begin
      t:= 1.0*j/5; // time must be between 0 and 1.
      t2:= sqr(t);
      t3:= t*t2;
      c1:= 2*t3-3*t2+1;
      c2:= t3-2*t2+t;
      c3:= -2*t3+3*t2;
      c4:= t3-t2;
      // This is the point in the section from i to i+1 for this t:
      x:= c1 * Points[i1,0] + c2*r[1,1] + c3*Points[i2,0] + c4*r[2,1];
      y:= c1 * Points[i1,1] + c2*r[1,2] + c3*Points[i2,1] + c4*r[2,2];
      // Detect the max. y value and keep both values:
      if (y > MaxY) then begin
        MaxY:= y;
        MaxX:= x;
      end;
      // Detect averaged decision score:
      if (OldX<=AvgValue) and (x>=AvgValue) then begin
        AvgScore:= y;
      end;
      OldX:= x;
    end;
  end;
end;

// Debugging:
procedure TListOfFutures.DrawOptions;
var f: integer;
begin
  for f:= 0 to FutureCount-1 do
    if not Future[f].FutureDiscarted then
      TPlayer(fOwner).DrawFutureScore(Future[f]);
  // Adjust scale of the circles:
  TPlayer(fOwner).ResetFutureScoreScale;
end;


end.
