unit UPlayer;

interface

uses Graphics, Types, Contnrs,
     UState, UListOfFutures, UGoal;

// Higher level of AI possible:
const Max_AI_Level : integer = 8;

type
  TPlayer = class(TObject)
  private
    // Team the player belongs to:
    fTeam: TObject;
    // Color for drawing:
    fColor: TColor;
    // Debug flags:
    fDebugTraces: boolean;
    fDebugOptions: boolean;
    fDebugGoals: boolean;
    fDebugLevels: boolean;
    // Debuging options' score is tricky, some times range from 0 to 1000.
    fFutureScoreScale, fFutureScoreMaxR: extended;
    // Am I imagining futures by taking random decisions?
    // When passing from false to true you must "initialize" FutureState, and
    // when setting it back to false, you must store the future found in your
    // internal list.
    fImagining: boolean;
    // Witch things motivate this player to move here or there?
    fGoals: TObjectList;
    // For testing purpouses you can downgrade AI to previous versions:
    // Level 1 : Future.FinalScore = 1 (except crashing futures that score 0)
    // Level 2 : Future.FinalScore = 1 always.
    // Level 3 : Future.FinalScore = 1 always, option's score = Ln(N).
    // Level 4 : Future.FinalScore = R (raced distance)
    // Level 5 : Future.FinalScore = R^2 (best formulae found so far)
    // Level 6 : 5 + Splines are used to help on Y-shaped situations (RefinedAI).
    // Level 7 : 5 + Lowest option's score is forced to be zero (ZeroedAI).
    // Level 8 : 7 + AutoAdjustSensitivity.
    fAI_Level: integer;
    // I duplicated my self in the last frame, flag it so goals can use it.
    fHadDuplicated: boolean;
    //
    // ************************
    // **  SETTER & GETTERS  **
    // ************************
    //
    // Actual state of the player (real or imaginary for futures):
    function GetState: TState;
    function GetTrack: TObject;
    function GetGoalCount: integer;
    // MicroState components:
    // ======================
    procedure SetDead(v: boolean);
    function  GetDead: boolean;
    function  GetOldDead: boolean;
    procedure SetSize(v: double);
    function  GetSize: double;
    function  GetMinSize: double;
    procedure SetEnergy(v: double);
    function  GetEnergy: double;
    function  GetOldEnergy: double;
    function  GetIniEnergy: double;
    procedure SetHealth(v: double);
    function  GetHealth: double;
    function  GetOldHealth: double;
    function  GetIniHealth: double;
    procedure SetImagining(v: boolean);
    function  GetGridSize: integer;
    procedure SetGridSize(v: integer);
    // Flags for AI_Level 6, 7 and 8
    function  GetUseRefinedAI: boolean;
    function  GetUseZeroedAI: boolean;
    function  GetAutoAdjustSensitivity: boolean;
    // Accesing player's default goals:
    function  GetGoalEnergy: TGoal;
    procedure SetGoalEnergyStrength(v: double);
    function  GetGoalEnergyStrength: double;
    function  GetGoalHealth: TGoal;
    procedure SetGoalHealthStrength(v: double);
    function  GetGoalHealthStrength: double;
  protected
    // A "State" contains all parameters (time-changing, fixed, AI driven) the player.
    // There are 2 ones: the actual state of the "real" player, and the "imaginary"
    // state used to imagine a future for this player. I will use "State" in the
    // code, but it will mask one of the two possible states depending on begin
    // really deciding with futures previously found, or you are precisely looking
    // for those futures.
    fActualState, fFutureState: TState;
    // Indexes of the StateParams that are common to all TPlayers:
    fSize, fDead, fEnergy, fHealth: integer;
    // Indexes of the Goals that are common to all TPlayers:
    fGoalEnergy: integer;
    fGoalHealth: integer;
    //
    // Grid size to convert from micro to macro state.
    // Computed params with a precision will change it as you redefine grid size.
    // If you define a target number of futures needed to decide properly, then
    // "AutoGridSize" will make grid size to change so the number of futures found
    // tend to fit the desired number (thinks better).
    fGridSize, fMaxGridSize, fMinGridSize: integer;
    fAutoGridSize: boolean;
    fTargetFutures: integer;
    NodesNeedIni: boolean;
    NodesNeedRecalc: boolean;
    function  DebugThisFuture: boolean; virtual;
    procedure CalcNodes; virtual; abstract;
    procedure DefineStateParams; virtual;
    function  GetGoal(m: integer): TGoal;
    function  GetSimMSeconds: integer;
    function  GetSimSeconds: double;
    function  GetCanDuplicate: boolean; virtual;
    property  FutureScoreScale: extended read fFutureScoreScale write fFutureScoreScale;
    property  FutureScoreMaxR:  extended read fFutureScoreMaxR  write fFutureScoreMaxR;
    // Evolution privates:
    function  MakeADuplicate: TPlayer; virtual; abstract;
  public
    // List of futures found:
    Futures: TListOfFutures;
    //
    constructor create;
    destructor  destroy; override;
    // Back to an initial state:
    procedure ResetState; virtual; //State  = Position, velocity... (change with time)
    // Add a new goal to the player's AI:
    procedure AddGoal(v: TGoal);
    procedure DelGoal(g: integer);
    // Copy params from another TPlayer:
    procedure CopyFrom(Other: TPlayer); virtual;
    // Drawing stuff:
    procedure Draw; virtual;
    function  DebugColor(ThisFuture: TState): TColor; virtual;
    procedure DrawFuturePoint; virtual; abstract;
    // Use goal strength to define color:
    procedure ColorUsingGoals;
    // Debugging end point scoring is hard, you need auto scale, so:
    procedure ResetFutureScoreScale;
    procedure DrawFutureScore(f: TState); virtual; abstract; // Level 3 debug
    // Grid size control:
    property GridSize: integer read GetGridSize write SetGridSize;
    property AutoGridSize: boolean read fAutoGridSize write fAutoGridSize;
    // After taking a decision, adjust grid size so number of futures found is ok:
    procedure AutoAdjustGridSize; virtual;
    //
    // Save actual state as the intial one:
    procedure SetIniState; virtual;
    property Team:  TObject read fTeam write fTeam;
    property Track: TObject read GetTrack;
    // Debug different things:
    property DebugTraces:  boolean read fDebugTraces  write fDebugTraces;
    property DebugOptions: boolean read fDebugOptions write fDebugOptions;
    property DebugGoals:   boolean read fDebugGoals   write fDebugGoals;
    property DebugLevels:  boolean read fDebugLevels  write fDebugLevels;
    // Color of the player:
    property Color: TColor read FColor write FColor;
    // SIMULATION:
    // ===========
    property  SimMSeconds: integer read GetSimMSeconds;
    property  SimSeconds: double read GetSimSeconds;
    procedure Simulate; virtual;
    procedure ScoreGoals(FireAt: TGoalFire);
    function  CheckCollision(Other: TPlayer): boolean; virtual; abstract;
    procedure BouncingDamage(bounce: double); virtual;
    // Witch is my state now? (the real or the imaginary one):
    property  State: TState read GetState;
    // Delete all futures as I am going to imagine a new set from my actual state:
    procedure ResetFutures;
    // Start/Stop imagining a new future involves managing the list of futures:
    property  Imagining: boolean read FImagining write SetImagining;
    // Taking decisions before moving:
    // Take the first random decision of a new future:
    procedure TakeFirstDecision;
    // Take another random decision to continue with a future:
    procedure TakeRandomDecision;
    // Take an intelligent decision based on futures found and their scorings:
    procedure TakeIntelligentDecision;
    // Get a goal by index number:
    property Goal[m: integer]: TGoal read GetGoal;
    property GoalCount: integer read GetGoalCount;
    // Debug it:
    procedure DrawOptions;
    // AI level (1 to 4 by now).
    property AI_Level: integer read fAI_Level write fAI_Level;
    property AutoAdjustSensitivity: boolean read GetAutoAdjustSensitivity;
    // Use refined decision layer to avoid Y shaped situations (AILevel=6).
    property UseRefinedAI: boolean read GetUseRefinedAI;
    // Use zeroed decision scores so decisions are stronger (AILevel=7).
    property UseZeroedAI: boolean read GetUseZeroedAI;
    property MinGridSize: integer read fMinGridSize write fMinGridSize;
    property MaxGridSize: integer read fMaxGridSize write fMaxGridSize;
    property Size: double read GetSize write SetSize;
    property MinSize: double read GetMinSize;
    property Dead: boolean read GetDead write SetDead;
    property OldDead: boolean read GetOldDead;
    property Energy: double read GetEnergy write SetEnergy;
    property OldEnergy: double read GetOldEnergy;
    property IniEnergy: double read GetIniEnergy;
    property OldHealth: double read GetOldHealth;
    property IniHealth: double read GetIniHealth;
    property Health: double read GetHealth write SetHealth;
    // Properties to manipulate the player's goals:
    property GoalEnergy: TGoal read GetGoalEnergy;
    property GoalEnergyStrength: double read GetGoalEnergyStrength write SetGoalEnergyStrength;
    property GoalHealth: TGoal read GetGoalHealth;
    property GoalHealthStrength: double read GetGoalHealthStrength write SetGoalHealthStrength;
    // Evolution:
    function  Clone: TPlayer; virtual;
    property  CanDuplicate: boolean read GetCanDuplicate;
    procedure CopyGoalsStrength(Other: TPlayer);
    property  HadDuplicated: boolean read fHadDuplicated write fHadDuplicated;
    procedure RandomizeGoals;
    procedure MutateGoals;
  end;

implementation

uses Math, UUtils, UTrack, UTeam, UGoalEnergy, UGoalHealth;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TPlayer.create;
begin
  // Create the underlying object:
  inherited;
  // State:
  fActualState:= TState.create(self);
  fFutureState:= TState.create(self);
  // Fill with all the components I need (position, angle, etc):
  DefineStateParams();
  // Futures:
  Futures:= TListOfFutures.create(self);
  // AI Level:
  AI_Level:= Max_AI_Level;
  // Grid size control:
  fGridSize:=    10;
  fMinGridSize:=  2;
  fMaxGridSize:= 20;
  fAutoGridSize:= true;
  fTargetFutures:= 100;
  // Set internal params:
  NodesNeedRecalc:= true;
  NodesNeedIni:= true;
  Imagining:= false;
  DebugTraces:= false;
  DebugOptions:= false;
  DebugGoals:= false;
  DebugLevels:= false;
  FutureScoreScale:= 1;
  fFutureScoreMaxR:= 0;
  // Set a random color:
  Color:= RGBToColor(100+round(random*155), 100+round(random*155), 100+round(random*155));
  SetIniState;
  // Personal goals in life:
  fGoals:= TObjectList.create;
  fGoals.OwnsObjects:= false;
  // Default set of goals for all players:
  //
  // Players have a natural strong (x1) "tendency" for keeping health high:
  AddGoal(TGoalHealth.create(1));
  fGoalHealth:= GoalCount-1;
  // Players have a natural weak (x0.1) "tendency" for keeping energy high:
  AddGoal(TGoalEnergy.create(0.1));
  fGoalEnergy:= GoalCount-1;
end;

destructor TPlayer.destroy;
begin
  Futures.Free;
  fActualState.Free;
  fFutureState.Free;
  while fGoals.Count > 0 do begin
    TGoal(fGoals.Items[0]).Free;
    fGoals.Delete(0);
  end;
  fGoals.Free;
  inherited;
end;

procedure TPlayer.CopyFrom(Other: TPlayer);
begin
  // Copy its State:
  fActualState.CopyFrom(Other.fActualState);
  // Reset initial state and scoring:
  SetIniState;
  // Thinking params:
  Self.Imagining:=    Other.Imagining; // Should be false.
  Self.AI_Level:=     Other.AI_Level;
  Self.AutoGridSize:= Other.AutoGridSize;
end;


// ******************
// ** STATES STUFF **
// ******************

procedure TPlayer.DefineStateParams;
begin
  // I will add items to fActualState, not fFutureState, so I set "Imagining" to
  // false (it should already be so) so State will point to fActualState when used
  // from a descendant's DefineStateParams.
  Imagining:= false;
  // Fixed params: (simulation don't change them):
  State.AddParam( TStateParam.createFixed(   'Size',     6,  25, 3) ); // Player's size
  State.AddParam( TStateParam.createComputed('Dead',     0,   1, 0) );
  State.AddParam( TStateParam.createComputed('Health',  50, 100, 0) );
  State.AddParam( TStateParam.createComputed('Energy',  50, 100, 0) ); // Internal energy
  // Index of all created params (for fast access as property):
  fSize:=   State.ParamIndexByName('Size');
  fDead:=   State.ParamIndexByName('Dead');
  fHealth:= State.ParamIndexByName('Health');
  fEnergy:= State.ParamIndexByName('Energy');
end;

// Get actual (real or imaginary) state depending on the proccess you are now:
// Moving or imagining new futures.
function TPlayer.GetState: TState;
begin
  if Imagining then
    result:= fFutureState
  else
    result:= fActualState;
end;

function TPlayer.GetTrack: TObject;
begin
  result:= TTeam(fTeam).Track;
end;

// Reset state to initial:
procedure TPlayer.ResetState;
begin
  NodesNeedRecalc:= true;
  fActualState.BackToIniState;
end;

// Set the actual state as the "initial":
procedure TPlayer.SetIniState;
begin
  // Reset scoring:
  State.ResetScore;
  // Copy actual values into ini and old parts of it.
  State.SetIniState();
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TPlayer.Draw;
begin
  // Before drawing you need to recalc node positions:
  CalcNodes;
  // Each kind of TPlayer will have to do the rest.
  // ...
end;

// Draw ending points of the futures:
procedure TPlayer.DrawOptions;
begin
  if DebugOptions then
    Futures.DrawOptions;
end;

// Adjust scale of the circles when drawing options/futures
procedure TPlayer.ResetFutureScoreScale;
const okradio: integer = 5;
begin
  // Change scale so max is inside +/- 20% of the okradio:
  if FutureScoreMaxR > (okradio*2) then
    FutureScoreScale:= Max(0.00001, FutureScoreScale / 2)
  else if (FutureScoreMaxR < (okradio/1.5)) then
    try FutureScoreScale:= FutureScoreScale * 2;
    finally end;
  // Clean the max:
  fFutureScoreMaxR:= 0;
end;

// Witch color do I use for debbugins this future? Red = left, etc.
function TPlayer.DebugColor(ThisFuture: TState): TColor;
begin
  result:= clGray;
end;

// May be you only want some futures to be shown while debugging:
function TPlayer.DebugThisFuture: boolean;
begin
  result:= true;
end;

// Use the goal strength to colorize players so evolution is visible on screen:
procedure TPlayer.ColorUsingGoals;
var
  i: integer;
  R: array [0..2] of integer;
begin
  for i:= 0 to 2 do begin
    if (GoalCount > i) then
      R[i]:= round(255*Goal[i].Strength/Goal[i].MaxStrength)
    else
      R[i]:= 128;
  end;
  Color:= RGBToColor(R[0], R[1], R[2]);
end;

// **********************
// ** SIMULATION STUFF **
// **********************

procedure TPlayer.Simulate;
begin
  // Time passes by!
  State.TimePassed:= State.TimePassed + SimMSeconds;
  // Each kind of TPlayer will have to do the rest...
  // ...
end;

// Simulation time (delta time):
function TPlayer.GetSimMSeconds: integer;
begin
  result:= TTrack(Track).SimMSeconds;
end;
function TPlayer.GetSimSeconds: double;
begin
  result:= TTrack(Track).SimSeconds;
end;

// If player bounce with track limits, calculate damage (-health):
procedure TPlayer.BouncingDamage(bounce: double);
var Damage: double;
begin
  // Standar way, bouncing hard is just bad for health:
  // Note: bounce is the velocity change (its length) occurred due to the bounce
  // so the energy associated is m*v^2, but mass is always 1 (some day... ).
  Damage:= Sqr(bounce/2);
  Health:= Health - Damage;
end;

// *****************
// ** GOALS STUFF **
// *****************

// After a simulation, the State is changed, and I need to score this change.
procedure TPlayer.ScoreGoals(FireAt: TGoalFire);
var g: integer;
begin
  for g:= 0 to GoalCount-1 do
    Goal[g].ScoreGoal(self, FireAt);
end;

// Add a new goal to the AI:
procedure TPlayer.AddGoal(v: TGoal);
begin
  fGoals.Add(v);
end;

// Delete a goal:
procedure TPlayer.DelGoal(g: integer);
begin
  if (g>=0) and (g<GoalCount) then
    fGoals.Delete(g);
end;

function TPlayer.GetGoal(m: integer): TGoal;
begin
  result:= TGoal(fGoals[m]);
end;

function TPlayer.GetGoalCount: integer;
begin
  result:= fGoals.count;
end;

// Accesing player's default goals:
// ================================

function TPlayer.GetGoalEnergy: TGoal;
begin
  result:= Goal[fGoalEnergy];
end;
function TPlayer.GetGoalEnergyStrength: double;
begin
  result:= GoalEnergy.Strength;
end;
procedure TPlayer.SetGoalEnergyStrength(v: double);
begin
  GoalEnergy.Strength:= v;
end;

function TPlayer.GetGoalHealth: TGoal;
begin
  result:= Goal[fGoalHealth];
end;
function TPlayer.GetGoalHealthStrength: double;
begin
  result:= GoalHealth.Strength;
end;
procedure TPlayer.SetGoalHealthStrength(v: double);
begin
  GoalHealth.Strength:= v;
end;

// ********************
// ** THINKING STUFF **
// ********************

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeFirstDecision;
begin
  // Before taking any step, store the old values:
  State.SetOldState;
  State.TakeFirstDecision;
end;

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeRandomDecision;
begin
  // Before taking any step, store the old values:
  State.SetOldState;
  State.TakeRandomDecision;
end;

// Take a decision before simulating the next delta time:
procedure TPlayer.TakeIntelligentDecision;
begin
  // Before taking any step, store the old values:
  State.SetOldState;
  // Analize futures found and take an intelligent decision based on them.
  Futures.TakeIntelligentDecision(State);
  // Auto adjust grid size to tend to a given number of diff. futures.:
  AutoAdjustGridSize;
end;

// Star/Stop imagining a new future triggers the creation and storing of it
procedure TPlayer.SetImagining(v: boolean);
begin
  // Detect only real changes:
  if (v <> fImagining) then begin
    // Starting to imagine a new future or just coming back to reality?
    if v then begin
      // Start of the imagination proccess of a new future:
      //
      // Basically you only need to set your "imaginary state" as the on/used
      // state as this will be the starting point for the future to come:
      fFutureState.CopyFrom(fActualState);
      // This is the initial state for this future:
      fFutureState.SetIniState;
      // Scoring of the future is still empty:
      fFutureState.ResetScore;
      // Flag it if you duplicated during this future:
      HadDuplicated:= false;
    end else begin
      // End of the imagination proccess of a new future:
      //
      // The goals that fires at the end of the future do it now:
      ScoreGoals(gfEnd);
      // The future is found, store it on my list of futures so I can later
      // decide witch was the best decision and take it on the actual state.
      Futures.AddFuture(fFutureState);
    end;
    // Set the new "mind state":
    fImagining:= v;
    // Nodes needs recalc as position has changed:
    NodesNeedRecalc:= true;
  end;
end;

// Delete all futures as I am going to imagine a new set from my actual state:
procedure TPlayer.ResetFutures;
begin
  Futures.Clear;
end;

// Use splines to avoid Y-shaped situations? (AI_Level 6)
// It is not from any entropy definition, so not a "genuine" evolution!
function TPlayer.GetUseRefinedAI: boolean;
begin
  result:= (AI_Level=6);
end;
// Lowest option score is forced to zero (AI_Level 7).
// "Entropy compliant" that alolows futures with negative scorings.
function TPlayer.GetUseZeroedAI: boolean;
begin
  result:= (AI_Level=7);
end;
// Adjust options width automatically (AI_Level 7).
// "Entropy compliant" that alolows futures with negative scorings.
function TPlayer.GetAutoAdjustSensitivity: boolean;
begin
  result:= (AI_Level=8);
end;



// *********************
// ** EVOLUTION STUFF **
// *********************

function TPlayer.GetCanDuplicate: boolean;
begin
  result:= false;
end;

procedure TPlayer.CopyGoalsStrength(Other: TPlayer);
var g, og: integer;
begin
  // For each of my goals:
  for g:= 0 to GoalCount-1 do begin
    // Find the similar one on the other's goals:
    for og:= 0 to Other.GoalCount-1 do begin
      if Goal[g].ClassName = Other.Goal[og].ClassName then begin
        // Copy strength:
        Goal[g].Strength:= Other.Goal[og].Strength;
        break;
      end;
    end;
  end;
end;

function TPlayer.Clone: TPlayer;
begin
  // First, duplicate the player:
  result:= MakeADuplicate;
  result.CopyGoalsStrength(self);
  HadDuplicated:= true;
  // Add it to my own team:
  TTeam(Team).AddPlayer(result);
  // Evolution tricks:
  if TTrack(Track).Evolution then begin
    // Share your energy with the new baby:
    self.Energy:= self.Energy/2;
    result.Energy:= self.Energy;
    // Change DNA a little (goal stenghts)
    MutateGoals;
    // Use 3 main goal strength to define player's color:
    result.ColorUsingGoals;
  end;
end;

// Randomly decide the goals' strengths:
procedure TPlayer.RandomizeGoals;
var g: integer;
begin
  for g:= 0 to GoalCount-1 do
    TGoal(Goal[g]).Strength:= (0.1+0.8*random) * TGoal(Goal[g]).MaxStrength;
end;

// Randomly change the goals' strengths:
procedure TPlayer.MutateGoals;
var
  g: integer;
const
  PercentChange: double = 0.1; // +/-10% of change.
begin
  for g:= 0 to GoalCount-1 do begin
    // Add or take a maximum of 10% of the strength range:
    Goal[g].Strength:= Goal[g].Strength +
      Goal[g].MaxStrength * PercentChange * (2 * random - 1);
  end;
end;


// ***********************
// ** GRID SIZE CONTROL **
// ***********************

// Grid size control:
function TPlayer.GetGridSize: integer;
begin
  result:= Min(fMaxGridSize, Max(fMinGridSize, fGridSize));
end;

// Modify the precision of computed params so converting from micro to macro
// state generate more or less different futures. If your player needs a more
// specific GridSize to Precision schema, override it, call inherited, then
// redefine the precisions as needed.
procedure TPlayer.SetGridSize(v: integer);
var i:integer;
begin
  fGridSize:= Min(fMaxGridSize, Max(fMinGridSize, v));
  // By default, all computed params with a precision (those that convert from
  // micro to macro state when a future is reached) copy this "GridSize" as its
  // "Precision", so if only PosX and PosY are defined, a grid size is exactly
  // what you should spect in a vectorial graphic editor.
  for i:= 0 to State.PositionalParamCount-1 do
    State.PositionalParam[i].Precision:= v;
end;

// After taking a decision, adjust grid size so the number of different futures
// tend to a "magical" number. This way, if your are about to crash and you get
// few futures, grid size decreases and futures are more "fine grained" so get
// get more of them and the decision is more stable:
procedure TPlayer.AutoAdjustGridSize;
begin
  // If I have to tend to get 100 futures, and found 50, then decrease grid size:
  // Do it by averaging last grid size with estimated one, so it changes smoothly.
  if AutoGridSize then
    GridSize:= round(GridSize * (1 + Futures.DifferentFutures / fTargetFutures)/2);
end;


// ******************************
// ** SET/GET FOR STATE PARAMS **
// ******************************

// Player's size:
procedure TPlayer.SetSize(v: double);
begin
  if (State.Param[fSize].Value <> v) then begin
    State.Param[fSize].Value:= v;
    NodesNeedRecalc:= true;
  end;
end;
function  TPlayer.GetSize: double;
begin
  result:= State.Param[fSize].Value;
end;
function  TPlayer.GetMinSize: double;
begin
  result:= State.Param[fSize].MinValue;
end;

// Dead
procedure TPlayer.SetDead(v: boolean);
begin
  if (v <> (State.Param[fDead].Value>0.5)) then begin
    if v then begin
      State.Param[fDead].Value:= 1;
      // Record the time of the death, Watson:
      State.TimeSurvived:= State.TimePassed;
    end else
      State.Param[fDead].Value:= 0;
  end;
end;
function  TPlayer.GetDead: boolean;
begin
  result:= (State.Param[fDead].Value>0.5);
end;
function  TPlayer.GetOldDead: boolean;
begin
  result:= (State.Param[fDead].OldValue>0.5);
end;

// Energy
procedure TPlayer.SetEnergy(v: double);
begin
  State.Param[fEnergy].Value:= v;
end;
function  TPlayer.GetEnergy: double;
begin
  result:= State.Param[fEnergy].Value;
end;
function  TPlayer.GetOldEnergy: double;
begin
  result:= State.Param[fEnergy].OldValue;
end;
function  TPlayer.GetIniEnergy: double;
begin
  result:= State.Param[fEnergy].IniValue;
end;

// Health = From 100 to 0, if 0 then you are dead!
procedure TPlayer.SetHealth(v: double);
begin
  State.Param[fHealth].Value:= v;
  Dead:= (State.Param[fHealth].Value = 0);
end;
function  TPlayer.GetHealth: double;
begin
  result:= State.Param[fHealth].Value;
end;
function  TPlayer.GetOldHealth: double;
begin
  result:= State.Param[fHealth].OldValue;
end;
function  TPlayer.GetIniHealth: double;
begin
  result:= State.Param[fHealth].IniValue;
end;

end.
