unit UPlayer2D;

interface

uses Types,
     UState, UPlayer, UGoal;

type
  TPlayer2D = class(TPlayer)
  private
    // Impact point with track limits:
    fImpacted: boolean;
    fImpactX, fImpactY: double;
    // State components
    // ================
    procedure SetPosX(v: double);
    function  GetPosX: double;
    procedure SetPosY(v: double);
    function  GetPosY: double;
    procedure SetAngle(v: double);
    function  GetAngle: double;
    function  GetOldAngle: double;
    function  GetOldPosX: double;
    function  GetOldPosY: double;
    procedure SetVelX(v: double);
    function  GetVelX: double;
    function  GetOldVelX: double;
    procedure SetVelY(v: double);
    function  GetVelY: double;
    function  GetOldVelY: double;
    procedure SetVelRot(v: double);
    function  GetVelRot: double;
    // Accesing player's default goals:
    function  GetGoalRaced: TGoal;
    procedure SetGoalRacedStrength(v: double);
    function  GetGoalRacedStrength: double;
  protected
    // Automatic update of vectors of the body:
    VectorsNeedRecalc: boolean;
    // Indexes of the StateParams that are common to all TPlayers:
    fPosX, fPosY, fAngle, fAngle2, fVelX, fVelY, fVelRot: integer;
    // Indexes of default goals:
    fGoalRaced: integer;
    function  GetMaxX: double;
    function  GetMinX: double;
    function  GetMaxY: double;
    function  GetMinY: double;
    procedure DefineStateParams; override;
    procedure CalcNodes; override;
    procedure CalcVectors;
    procedure MemorizePosition;
    // Evolution:
    function  GetCanDuplicate: boolean; override;
  public
    // Those 2D node list forms the body polygon:
    NodesR: array of array[1..2] of double;
    Nodes: array of TPoint;
    // Vectors connecting node i with next node:
    Vectors: array of array[1..2] of double;
    // Administrative:
    constructor create;
    // Change ini and actual 2D position:
    procedure PlaceHere(x, y, a: double);
    // Move to a new relative 2D position:
    procedure HandMove(dx, dy, da: double; repaint: boolean = true);
    // Simulation:
    procedure Simulate; override;
    function  IsInsideImage: boolean;
    procedure IsInsideTrack; // It will set "Impacted" and (ImpactX,  ImpactY)
    function  IsInsidePlayer(x, y: double): boolean;
    procedure BounceWithFence;
    function  CheckCollision(Other: TPlayer): boolean; override;
    // Draw the player on the track:
    procedure Draw; override;
    procedure DrawBoxLimits;
    procedure DrawBigPoint;
    // Draw debug information:
    procedure DrawFuturePoint; override;
    procedure DrawFutureScore(f: TState); override;
    // 2D box containg the player:
    property MaxX: double read GetMaxX;
    property MinX: double read GetMinX;
    property MaxY: double read GetMaxY;
    property MinY: double read GetMinY;
    // Impact point:
    property Impacted: boolean read fImpacted;
    property ImpactX: double read fImpactX;
    property ImpactY: double read fImpactY;
    // State params:
    property PosX:    double read GetPosX   write SetPosX;
    property PosY:    double read GetPosY   write SetPosY;
    property Angle:   double read GetAngle  write SetAngle;
    property VelX:    double read GetVelX   write SetVelX;
    property VelY:    double read GetVelY   write SetVelY;
    property VelRot:  double read GetVelRot write SetVelRot;
    // Previous state params:
    property OldPosX:  double read GetOldPosX;
    property OldPosY:  double read GetOldPosY;
    property OldAngle: double read GetOldAngle;
    property OldVelX:  double read GetOldVelX;
    property OldVelY:  double read GetOldVelY;
    // Properties to manipulate the player's goals:
    property GoalRaced: TGoal read GetGoalRaced;
    property GoalRacedStrength: double read GetGoalRacedStrength write SetGoalRacedStrength;
  end;

implementation


uses Graphics, Math, Forms, SysUtils,
     UUTils, UTrack2D, UGoal2DRaced, UGoalTimer, UGoalDuplicate, UTrack;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TPlayer2D.create;
begin
  inherited;
  // Set an intial position and angle:
  PosX:= 10;
  PosY:= 10;
  Angle:= 0;
  fTargetFutures:= 100;
  // 2D players have a natural "tendency" for racing fast and inside the track:
  AddGoal(TGoal2DRaced.create(1));
  fGoalRaced:= GoalCount-1;
  // 2D player duplicate if evolution is ON:
  AddGoal(TGoalDuplicate.create(0.5));
end;

procedure TPlayer2D.DefineStateParams;
begin
  inherited;
  // A 2D TPlayer must have a PosX, PosY Angle and its velocities, at least:
  State.AddParam( TStateParam.createPositional('PosX',  0.0, 15.0) );
  State.AddParam( TStateParam.createPositional('PosY',  0.0, 15.0) );
  //State.AddParam( TStateParam.createPositional('Angle2',0.0, 15.0) );
  State.AddParam( TStateParam.createComputed('Angle2',0.0, 0.0, 0.0) );
  // Angle is defined as computed so not used in macro states, you could opt differently!
  State.AddParam( TStateParam.createComputed('Angle', 0.0, 0.0, 0.0) );
  State.AddParam( TStateParam.createComputed('VelX',  0.0, 0.0, 0.0) );
  State.AddParam( TStateParam.createComputed('VelY',  0.0, 0.0, 0.0) );
  State.AddParam( TStateParam.createComputed('VelRot',0.0, 0.0, 0.0) );
  // Index of all created params (for fast access as property):
  fPosX:=  State.ParamIndexByName('PosX');
  fPosY:=  State.ParamIndexByName('PosY');
  fAngle:= State.ParamIndexByName('Angle');
  fAngle2:=State.ParamIndexByName('Angle2');
  fVelX:=  State.ParamIndexByName('VelX');
  fVelY:=  State.ParamIndexByName('VelY');
  fVelRot:=State.ParamIndexByName('VelRot');
end;


// **********************
// ** SIMULATION STUFF **
// **********************

procedure TPlayer2D.Simulate;
begin
  // Store old values, count time passing by...
  inherited;
  // OFF-TRACK CONTROL
  // =================
  // Are you crashing with track limits? (impacted=true)
  IsInsideTrack;
  // Leaving the screen is really bad, avoid it or die:
  if not IsInsideImage then
    dead:= true;
  // In order to be able to test old AI levels I need to add a death penalty
  // if you crash the player, as it was mean to happen with levels 1 to 3:
  if (AI_Level <= 3) then begin
    // This will stop player from running in a future after a crash, it will just
    // stop and the raced distance will end in the impact point:
    if (Health<1) and Imagining then begin
      VelX:=   0;
      VelY:=   0;
      VelRot:= 0;
    end;
  end;

  // Dead players lose size and when it gets to zero, they pass!
  if Dead and not Imagining then begin
    Size:= Size - SimSeconds * 0.5; // Lose 1 unit of size each 2 seconds.
  end;

  // The rest is in the player's simulation code...
end;

// Bounce back on the fence of the track if you try to go off-track:
procedure TPlayer2D.BounceWithFence;
var
  i: integer;
  // Vars to react from a impact on a node of the player's body:
  C2I_x, C2I_y, dist: double; // Vector Player's center -> PointOfImpact
  C2In_x, C2In_y: double;     // Same I2C but normalized to length 1.
  RF_x, RF_y, RF_l: double;   // Reaction vel. = -Velocity of the point of impact.
  Cross, Prod: double;
  // Small 1/10 steps to the last inside position:
  ActX, ActY, ActA: double; //  change variable name! And RF is not a force, is a velocity
  // Accumulate all vels. that will be added to player's state: you can not add
  // them on each impact point found as it would interfer with the next impact
  // point you find and will over react to the first impact point:
  AddVelX, AddVelY, AddVelRot: double;
const
  Elasticity: double = 0.9; // 1 for 100% bouncing, 0 for absorb all the impacts
begin
  if not Impacted then exit;
  // Walk back in time to the impact time (in 1/10 steps):
  // Note1: If you crashed too hard, you can have more than half of the player
  // outside. Using center point to bounce will work "reversed".
  // Note2: After calculating reactions, a final 1/10 step must be done in order
  // to let the player inside the track.
  ActX:= (State.Param[fPosX].MemoValue  - PosX)  / 10;
  ActY:= (State.Param[fPosY].MemoValue  - PosY)  / 10;
  ActA:= (State.Param[fAngle].MemoValue - Angle) / 10;
  i:= 10;
  repeat
    PosX:=  PosX + ActX;
    PosY:=  PosY + ActY;
    Angle:= Angle + ActA;
    dec(i);
    IsInsideTrack;
    if not Impacted then begin
      // OK, I did an extra step, so undo it:
      PosX:= PosX  - ActX;
      PosY:= PosY  - ActY;
      Angle:=Angle - ActA;
      IsInsideTrack;
      i:= 0;
    end;
  until (i<=0);
  // I will accumulate the "reactinons" of the player to the impact points:
  AddVelX:= 0; AddVelY:= 0; AddVelRot:= 0;

  {
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
  TTrack2D(Track).Img.Canvas.Pen.Color:= clRed;
  TTrack2D(Track).Img.Canvas.Ellipse(round(ImpactX-2), round(ImpactY-2), round(ImpactX+2), round(ImpactY+2));
  Application.ProcessMessages;
  }

  // I2C = Vector from impact point to the center of the player:
  C2I_x:= ImpactX - PosX;
  C2I_y:= ImpactY - PosY;
  // Well, I really will need it normalized to length 1 later:
  dist:= Sqrt(Sqr(C2I_x)+Sqr(C2I_y));
  // Impact point = center point? Not possible, but...
  if (dist=0) then exit;
  // Normalize it:
  C2In_x:= C2I_x / dist;
  C2In_y:= C2I_y / dist;
  //
  // Real velocity of the impact point: Player's vel. + Rotational vel.
  // Note1: Normal vector used (-C2I_y, +C2I_x), Rot.Vel = Normal * VelRot(rad/s).
  // Note2: The reaction force of the ground is -VReal, so I change sings.
  RF_x:= -(VelX - C2I_y * VelRot);
  RF_y:= -(VelY + C2I_x * VelRot);

  {
  TTrack2D(Track).Img.Canvas.Pen.Color:= clRed;
  TTrack2D(Track).Img.Canvas.MoveTo(round(ImpactX), round(ImpactY));
  TTrack2D(Track).Img.Canvas.LineTo(round(ImpactX+RF_x), round(ImpactY+RF_y));
  Application.ProcessMessages;
  }

  // RF_L = Damage to the player!
  RF_l:= Sqrt(Sqr(RF_x)+Sqr(RF_y));
  // Health:= Health - RF_l;

  // Change the player's velocity with bounce = FR proyected over I2C:
  Cross:= ((RF_x * C2In_X) + (RF_y * C2In_Y));
  AddVelX:= AddVelX + (Cross * C2In_x);
  AddVelY:= AddVelY + (Cross * C2In_y);

  {
  TTrack2D(Track).Img.Canvas.Pen.Color:= clBlue;
  TTrack2D(Track).Img.Canvas.MoveTo(round(PosX), round(PosY));
  TTrack2D(Track).Img.Canvas.LineTo(round(PosX+AddVelX), round(PosY+AddVelY));
  Application.ProcessMessages;
  }

  // Rest of the reaction force? (Must be perpendicular to I2C)
  RF_x:= RF_x - (Cross * C2I_x);
  RF_y:= RF_y - (Cross * C2I_y);
  // Length of N (normal velocity) / radio = angular velocity to add:
  // Note: N is the z-coordinte on vectorial product (C2In ^ RV).
  Prod:= (C2In_x * RF_Y) - (C2In_y * RF_X);
  AddVelRot:= AddVelRot + (Prod / dist);

  // Correct velocity of the player as a reaction to all the impact points found:
  VelX:=   Elasticity * (VelX   + 2*AddVelX);
  VelY:=   Elasticity * (VelY   + 2*AddVelY);
  VelRot:= Elasticity * AddVelRot; // Elasticity * (VelRot + 2*AddVelRot);
  // Now that the velocity has been corrected, separate player from fence:
  //
  // Step inside track before concluding the "bounce" proccess:
  PosX:=  PosX + ActX;
  PosY:=  PosY + ActY;
  Angle:= Angle + ActA;
  IsInsideTrack;
  // Take health away after crash:
  BouncingDamage(RF_l);
end;

function TPlayer2D.IsInsideImage: boolean;
var i: integer;
begin
  CalcNodes;
  // Ask the Track if the player's node positions are inside image.
  for i:= Low(Nodes) to High(Nodes) do begin
    if not TTrack2D(Track).IsInsideImage(Nodes[i].x, Nodes[i].y) then begin
      result:= false;
      exit;
    end;
  end;
  result:= true;
end;

procedure TPlayer2D.IsInsideTrack;
var
  i: integer;
  nImp: integer;
begin
  CalcNodes;
  NImp:= 0;
  // Ask the Track if the player is on the track or outside it.
  for i:= Low(Nodes) to High(Nodes) do begin
    // If one point is off-track, then player is off-track:
    if not TTrack2d(Track).IsInsideTrack(Nodes[i].x, Nodes[i].y) then begin
      if (NImp=0) then begin
        // This way impact will be available even after the impact.
        fImpactX:= 0;
        fImpactY:= 0;
      end;
      fImpactX:= fImpactX + NodesR[i, 1];
      fImpactY:= fImpactY + NodesR[i, 2];
      inc(NImp);
    end;
  end;
  if (NImp>0) then begin
    // Average impact points:
    fImpactX:= fImpactX / NImp;
    fImpactY:= fImpactY / NImp;
  end else begin
    // Memorize position: "last known inside position"
    MemorizePosition;
  end;
  fImpacted:= (NImp>0);
end;

// Is the point (x,y) inside the player's polygon?
function TPlayer2D.IsInsidePlayer(x, y: double): boolean;
var
  OK_Side, Side, i: integer;
  Cross: double;
begin
  // From my nodes, build the vectors that connect node i to i+1:
  CalcVectors;
  // Point (x,y) is inside of a convex poligon "B->C->D->E" if and only if all
  // cross-product (B->C)^(B->A) (with B->C a vector of the poligon) has the same
  // sign (i.e. point "A" is always at the right -or left- side of the vectors).
  OK_Side:= 0;
  result:= true;
  for i:= 0 to High(Nodes) do begin
    Cross:= (x - Nodes[i].x) * Vectors[i, 2] - (y - Nodes[i].y) * Vectors[i,1];
    // Point A is in one side of vector B->C of the polygon, remember the side:
    Side:= Sign(Cross);
    // 0 means point A is aligned with vector B->C, just on the border, so
    // don't count it as a prove it is out or in, just forget about this.
    if (Side<>0) then begin
      if (OK_Side=0) then
        // First product non-cero defines the "side" at witch point "A" must
        // always be to be inside the polygon.
        OK_Side:= Side
      else if (OK_Side <> Side) then begin
        // "A" is not at the right side of this vector, so it is outside:
        result:= false;
        break;
      end;
    end;
  end;
end;

// 2 Players collide when a node of one is inside of the other's poligon.
// Point inside a convex poligon? All cross-products must be same signed.
function TPlayer2D.CheckCollision(Other: TPlayer): boolean;
var
  Other2D: TPlayer2D;
  i: integer;
begin
  result:= false;
  // Only Player2D can be checked:
  if (Other is TPlayer2D) then
    Other2D:= TPlayer2D(Other)
  else
    exit;
  // Force node recalculation if necessary:
  CalcNodes;
  Other2D.CalcNodes;
  // If ONE of the other's nodes is inside me, return true and exit:
  for i:= 0 to High(Other2D.Nodes) do begin
    if IsInsidePlayer(Other2D.Nodes[i].x, Other2D.Nodes[i].y) then begin
      result:= true;
      exit;
    end;
  end;
end;

// Set the inital "position" manually:
procedure TPlayer2D.PlaceHere(x, y, a: double);
begin
  // Set position:
  PosX:=  x;
  PosY:=  y;
  Angle:= a;
  // Stop the movement:
  VelX:= 0;
  VelY:= 0;
  VelRot:= 0;
  // Use it as the "initial" position:
  SetIniState;
end;

// Move the player from here:
procedure TPlayer2D.HandMove(dx, dy, da: double; repaint: boolean = true);
begin
  // Update my state:
  PosX:=  PosX  + dx;
  PosY:=  PosY  + dy;
  Angle:= Angle + da;
  // Repaint?
  if repaint and Assigned(Track) then
    TTrack2D(Track).Draw(true);
end;

// Take a snapshot of the actual position:
procedure TPlayer2D.MemorizePosition;
begin
  State.Param[fPosX].MemorizeValue;
  State.Param[fPosY].MemorizeValue;
  State.Param[fAngle].MemorizeValue;
end;

// *******************
// ** DRAWING STUFF **
// *******************

procedure TPlayer2D.Draw;
var
  C: TCanvas;

  procedure DrawLevel(value: double; desp: integer);
  begin
    if (value > 0) then begin
      C.Pen.Color:= RGBToColor(150,150,150); // Grey
      C.Brush.Color:= clWhite;
      C.Rectangle(round(PosX+Size*1.3+desp), round(PosY+Size+1),
                  round(PosX+Size*1.3+desp+4), round(PosY-Size-1));
      // Filled up to:
      if (value>80) then
        C.Brush.Color:= clGreen
      else if (value>60) then
        C.Brush.Color:= clYellow
      else if (value>40) then
        C.Brush.Color:= RGBToColor(255,150,100) //Orange
      else
        C.Brush.Color:= clRed;
      C.Rectangle(round(PosX+Size*1.3+desp), round(PosY+Size),
                  round(PosX+Size*1.3+4+desp), round(PosY+Size-round(2*Size*value/100)));
    end;
  end;

begin
  if not Assigned(Track) then exit;
  inherited;
  CalcNodes;
  C:= TTrack2D(Track).Img.Canvas;
  // Draw the player's body:
  C.Pen.Color:= clBlack;
  C.Brush.Color:= Self.Color;
  C.Polygon(Nodes);
  if not dead then begin
    // Energy level is always shown:
    DrawLevel(Energy, 0);
    // Health level:
    DrawLevel(Health, 3);
    if DebugLevels then begin
      // To shows some value as text:
      //C.Brush.Style:= bsClear;
      //C.TextOut(round(PosX+Size*1.0), round(PosY-2.5*Size), Format('%3.2f', [Value here]) );
      // Sensitivity (of the first free param only) level:
      if AutoAdjustSensitivity and (State.FreeParamCount > 0) then
        DrawLevel(State.FreeParam[0].Sensitivity * 100, 10);
    end;
  end else begin
    C.MoveTo(round(PosX-Size), round(PosY-Size));
    C.LineTo(round(PosX+Size), round(PosY+Size));
    C.MoveTo(round(PosX-Size), round(PosY+Size));
    C.LineTo(round(PosX+Size), round(PosY-Size));
  end;
end;

// Draw the 2D bounding box (debuging):
procedure TPlayer2D.DrawBoxLimits;
begin
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
  TTrack2D(Track).Img.Canvas.Pen.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Rectangle(round(MinX), round(MinY), round(MaxX), round(MaxY));
end;
// Draw a big point to remark a position (debugging):
procedure TPlayer2D.DrawBigPoint;
const
  ra: integer = 2;
begin
  TTrack2D(Track).Img.Canvas.Brush.Style:= bsSolid;
  TTrack2D(Track).Img.Canvas.Pen.Color:= clBlack;
  TTrack2D(Track).Img.Canvas.Brush.Color:= Self.Color;
  TTrack2D(Track).Img.Canvas.Ellipse(round(PosX-ra), round(PosY-ra), round(PosX+ra), round(PosY+ra));
end;

// Nodes are the points that form the silouette of the player, as a base, I
// will assume it is a triangle, but each descendant of TPLayer must shape
// its nodes to look ok on screen
procedure TPlayer2D.CalcNodes;
var
  i: integer;
const
  Angles: array[0..2] of double = (0.0, 0.8, -0.8);
begin
  // Nodes is a dynamic array, so I must define size on first run:
  if NodesNeedIni then begin
    SetLength(Nodes, 3);
    SetLength(NodesR, 3);
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= Low(Nodes) to High(Nodes) do begin
      NodesR[i,1]:= PosX + Size * cos(Angle + Angles[i]*PI);
      NodesR[i,2]:= PosY + Size * sin(Angle + Angles[i]*PI);
      Nodes[i].X:= round(NodesR[i,1]);
      Nodes[i].Y:= round(NodesR[i,2]);
    end;
    NodesNeedRecalc:=   false;
    VectorsNeedRecalc:= true;
  end;
end;

procedure TPlayer2D.CalcVectors;
var
  NextI, i: integer;
begin
  CalcNodes;
  if VectorsNeedRecalc then begin
    VectorsNeedRecalc:= false;
    // From my nodes, build the vectors that connect node i to i+1:
    SetLength(Vectors, 1+High(Nodes));
    for i:= 0 to High(Nodes) do begin
      if (i=High(Nodes)) then
        NextI:= 0
      else
        NextI:= i+1;
      Vectors[i, 1]:= (NodesR[NextI, 1] - NodesR[i, 1]);
      Vectors[i, 2]:= (NodesR[NextI, 2] - NodesR[i, 2]);
    end;
  end;
end;

// Draw trace of the future (debug):
procedure TPlayer2D.DrawFuturePoint;
begin
  inherited;
  // Some futures must be hidden at debbug time:
  if DebugThisFuture then begin
    TTrack2D(Track).Img.Canvas.Pen.Color:= DebugColor(State);
    TTrack2D(Track).Img.Canvas.MoveTo(round(OldPosX), round(OldPosY));
    TTrack2D(Track).Img.Canvas.LineTo(round(PosX),    round(PosY));
  end;
end;

// Draw circles around a future ending point with radio = scoring.
procedure TPlayer2D.DrawFutureScore(f: TState);
var
  rr: double;
  x, y, r: integer;
  s: double;
begin
  x:= round(f.ParamByName['PosX'].Value);
  y:= round(f.ParamByName['PosY'].Value);
  // Score of the future to draw:
  s:= f.Score;
  rr:= Min(100.0, Abs(s) * 1.0 * FutureScoreScale);
  r:= round(rr+1);
  FutureScoreMaxR:= Max(FutureScoreMaxR, rr);
  // Draw the circle:
  if (s>=0) then begin
    // positive score draws circles:
    TTrack2D(Track).Img.Canvas.Pen.Color:= DebugColor(f);
    TTrack2D(Track).Img.Canvas.Brush.Style:= bsClear;
    TTrack2D(Track).Img.Canvas.Ellipse(x-r, y-r, x+r, y+r);
  end else begin
    // If score was negative, draw a black/white square:
    TTrack2D(Track).Img.Canvas.Pen.Color:=   clWhite;
    TTrack2D(Track).Img.Canvas.Brush.Color:= clBlack;
    TTrack2D(Track).Img.Canvas.Rectangle(x-r, y-r, x+r, y+r);
  end;
end;

// 2D box containing player:
function  TPlayer2D.GetMaxX: double;
begin
  result:= PosX+Size;
end;
function  TPlayer2D.GetMinX: double;
begin
  result:= PosX-Size;
end;
function  TPlayer2D.GetMaxY: double;
begin
  result:= PosY+Size;
end;
function  TPlayer2D.GetMinY: double;
begin
  result:= PosY-Size;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

function TPlayer2D.GetCanDuplicate: boolean;
begin
  result:= (Energy>=99);
end;



// *******************
// ** SET/GET STUFF **
// *******************

procedure TPlayer2D.SetPosX(v: double);
begin
  State.Param[fPosX].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetPosX: double;
begin
  result:= State.Param[fPosX].Value;
end;

procedure TPlayer2D.SetPosY(v: double);
begin
  State.Param[fPosY].Value:= v;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetPosY: double;
begin
  result:= State.Param[fPosY].Value;
end;

procedure TPlayer2D.SetAngle(v: double);
begin
  State.Param[fAngle].Value:= v;
  State.Param[fAngle2].Value:= round(v*180/PI) mod 360;
  NodesNeedRecalc:= true;
end;
function  TPlayer2D.GetAngle: double;
begin
  result:= State.Param[fAngle].Value;
end;

procedure TPlayer2D.SetVelX(v: double);
begin
  State.Param[fVelX].Value:= v;
end;
function  TPlayer2D.GetVelX: double;
begin
  result:= State.Param[fVelX].Value;
end;
function  TPlayer2D.GetOldVelX: double;
begin
  result:= State.Param[fVelX].OldValue;
end;

procedure TPlayer2D.SetVelY(v: double);
begin
  State.Param[fVelY].Value:= v;
end;
function  TPlayer2D.GetVelY: double;
begin
  result:= State.Param[fVelY].Value;
end;
function  TPlayer2D.GetOldVelY: double;
begin
  result:= State.Param[fVelY].OldValue;
end;

procedure TPlayer2D.SetVelRot(v: double);
begin
  State.Param[fVelRot].Value:= v;
end;
function  TPlayer2D.GetVelRot: double;
begin
  result:= State.Param[fVelRot].Value;
end;

// Old values:
function TPlayer2D.GetOldPosX: double;
begin
  result:= State.Param[fPosX].OldValue;
end;

function  TPlayer2D.GetOldPosY: double;
begin
  result:= State.Param[fPosY].OldValue;
end;

function  TPlayer2D.GetOldAngle: double;
begin
  result:= State.Param[fAngle].OldValue;
end;

// Accesing player's default goals:
// ===============================
function TPlayer2D.GetGoalRaced: TGoal;
begin
  result:= Goal[fGoalRaced];
end;
function TPlayer2D.GetGoalRacedStrength: double;
begin
  result:= GoalRaced.Strength;
end;
procedure TPlayer2D.SetGoalRacedStrength(v: double);
begin
  GoalRaced.Strength:= v;
end;


end.
