unit URocket;

interface

uses
  ExtCtrls, Graphics, Math, Types, DateUtils, SysUtils, Contnrs, Forms,
  UTrack2D, UState, UPlayer, UPlayer2D, UUtils;

type
  TRocket = class(TPlayer2D)
  private
    // State params' indexes:
    fRotaPower, fMainPower, fRotaPush, fMainPush: integer;
    // State components set/get:
    procedure SetRotaPower(v: double);
    function  GetRotaPower: double;
    procedure SetMainPower(v: double);
    function  GetMainPower: double;
    procedure SetRotaPush(v: double);
    function  GetRotaPush: double;
    procedure SetMainPush(v: double);
    function  GetMainPush: double;
  protected
    // Define your state components:
    procedure DefineStateParams; override;
    // Update the points on screen that form the player's siluette:
    procedure CalcNodes; override;
    // I only want left/right futures to be shown:
    function  DebugThisFuture: boolean; override;
    // Evolution:
    function  MakeADuplicate: TPlayer; override;
  public
    constructor create(x, y, a: double);
    procedure Simulate; override;
    procedure Draw; override;
    function  DebugColor(ThisFuture: TState): TColor; override;
    procedure BouncingDamage(bounce: double); override;
    // State params:
    property  MainPower: double read GetMainPower write SetMainPower;
    property  RotaPower: double read GetRotaPower write SetRotaPower;
    property  MainPush:  double read GetMainPush  write SetMainPush;
    property  RotaPush:  double read GetRotaPush  write SetRotaPush;
  end;


implementation

uses UTrack, UTeam;


// ****************************
// **  ADMINISTRATIVE STUFF  **
// ****************************

constructor TRocket.create(x, y, a: double);
begin
  inherited create;
  PlaceHere(x, y, a);
  Size:= 12;
  fTargetFutures:= 250;
  SetIniState;
end;

procedure TRocket.DefineStateParams;
begin
  // PosX and PosY are mandatory, so they are defined on TPLayer2D:
  inherited;
  // Add state params: (Name, Type, Value, Max, Min, Precision)
  // Adjusted for a friction 0.95 and delta time 0.1 (bad physics if it need adjusting):
  State.AddParam(TStateParam.createComputed('MainPower', 0, 50,   0));
  State.AddParam(TStateParam.createComputed('RotaPower', 0,  3,  -3));
  State.AddParam(TStateParam.createFree('MainPush', 80));
  State.AddParam(TStateParam.createFree('RotaPush',  15));
  // State param's Indexes:
  fMainPower:= State.ParamIndexByName('MainPower');
  fRotaPower:= State.ParamIndexByName('RotaPower');
  fMainPush:=  State.ParamIndexByName('MainPush');
  fRotaPush:=  State.ParamIndexByName('RotaPush');
end;


// *********************
// **  DRAWING STUFF  **
// *********************

procedure TRocket.CalcNodes;
var
  i: integer;
  r, a: double;
// Polar coordinates for the body points:
//const Angl: array[0..6] of double = (0, 26.6, 135, 135, 225, 225, 333);
//const Dist: array[0..6] of double = (3, 2.24, 1.41, 2.83, 2.83, 1.41, 2.24);
const Angl: array[0..4] of double = (0, 26.6, 135, 225, 333);
const Dist: array[0..4] of double = (3, 2.24, 2.83, 2.83, 2.24);
begin
  // Nodes is a dynamic array, so I must set size on first use.
  if NodesNeedIni then begin
    SetLength(Nodes,  1+High(Angl));
    SetLength(NodesR, 1+High(Angl));
    NodesNeedIni:= false;
    NodesNeedRecalc:= true;
  end;
  // If nothing changed from last time, don't waste time:
  if NodesNeedRecalc then begin
    for i:= 0 to High(Angl) do begin
      r:= Dist[i]/3*Size;
      a:= Angl[i]/180*PI;
      NodesR[i, 1]:= PosX + r * cos(Angle + a);
      NodesR[i, 2]:= PosY + r * sin(Angle + a);
      Nodes[i].X:= round(NodesR[i, 1]);
      Nodes[i].Y:= round(NodesR[i, 2]);
    end;
    NodesNeedRecalc:= false;
    VectorsNeedRecalc:= true;
  end;
end;


procedure TRocket.draw;
var
  C: TCanvas;
  IniX, IniY, EndX, EndY: integer;
const WinSize: double = 0.3;
begin
  // TPlayer will draw the body for me using the nodes:
  inherited;
  C:= TTrack2D(Track).Img.Canvas;
  // Window for astronuats:
  C.Brush.Color:= clSkyBlue;
  C.Pen.Color:= clBlue;
  C.Ellipse(round(PosX-WinSize*Size), round(PosY-WinSize*Size), round(PosX+WinSize*Size), round(PosY+WinSize*Size));
  // Lower point of the rocket = flames come from there
  IniX:= round( PosX - Cos(Angle) * Size );
  IniY:= round( PosY - Sin(Angle) * Size );
  // Draw main power:
  C.Pen.Color:= clRed;
  EndX:= IniX - round( Cos(Angle) * MainPower );
  EndY:= IniY - round( Sin(Angle) * MainPower );
  C.MoveTo(IniX, IniY);
  C.LineTo(EndX, EndY);
  // Draw left/right rotational power:
  EndX:= IniX + round( cos(Angle+PI/2) * (RotaPower) * 10 );
  EndY:= IniY + round( sin(Angle+PI/2) * (RotaPower) * 10 );
  C.MoveTo(IniX, IniY);
  C.LineTo(EndX, EndY);
end;

// Debug only left/right decision, not run/stop ones:
function TRocket.DebugThisFuture: boolean;
begin
  result:= (State.Param[fRotaPush].IniValue <> 0);
end;

// Debug lines are blue/red for left/right decisions:
function TRocket.DebugColor(ThisFuture: TState): TColor;
var C1, C2: integer;
begin
  // Red or blue color for left/right push:
  C1:= 255;
  // As future goes away it fades to white:
  C2:= Round( 155 + 100 * ThisFuture.TimePassed / TTrack(Track).FutureMSeconds);
  if (ThisFuture.Param[fRotaPush].IniValue > 0) then
    // Right turning is red:
    result:= RGBToColor(C1, C2, C2)
  else
    // Left turning is blue:
    result:= RGBToColor(C2, C2, C1);
end;


// ***********************
// **  SIMULATION STUFF **
// ***********************

// Let run the simulation for a delta time and update state acordingly:
procedure TRocket.Simulate;
var
  s, s2, g: double;
  AccX, AccY, Friction: double;
  UseMainPower, UseRotaPower: double;
begin
  // Inherited will store old values for me:
  inherited;
  // Milliseconds to seconds:
  s:= SimSeconds;
  s2:= Sqr(s)/2;
  g:= 9.8;
  Friction:= 0.99; // before 0.96
  // Push joystick => Power
  if Dead then begin
    MainPower:= 0;
    RotaPower:= 0;
  end else begin
    MainPower:= MainPower + MainPush * s;
    RotaPower:= RotaPower + RotaPush * s;
  end;
  // Power => depletes energy:
  MainPower:= Min(Energy/s*40, MainPower);
  Energy:= Max(0, Energy - MainPower*s/40);
  RotaPower:= Sign(RotaPower) * Min(Energy/s*10, Abs(RotaPower));
  Energy:= Max(0, Energy - Abs(RotaPower)*s/10);
  // The power available may be less than the power required...
  UseMainPower:= MainPower;
  UseRotaPower:= RotaPower;
  // Power => acceleration
  AccX:= UseMainPower * Cos(Angle);
  AccY:= UseMainPower * Sin(Angle) + g; // g=gravity
  // Acceleration => velocity:
  VelRot:= VelRot + UseRotaPower * s;
  VelX:=   VelX   + AccX * s;
  VelY:=   VelY   + AccY * s;
  // Velocity => position:
  Angle:=  Angle + VelRot * s;
  PosX:=   PosX + VelX * s + AccX * s2;
  PosY:=   PosY + VelY * s + AccY * s2;
  // Apply friction:
  VelRot:= VelRot * 0.90; // Too low friction causes mad spinings!
  VelX:=   VelX   * Friction;
  VelY:=   VelY   * Friction;
end;

// If rocket bounce with track limits, calculate damage, but may be it landed OK:
procedure TRocket.BouncingDamage(bounce: double);
var Damage, Coef, Absorb, ImpX, ImpY: double;
begin
  // This is the standard way:
  Damage:= Sqr(bounce/2) + (10*Sqr(VelRot));
  // If contac point is downside, then you can absorb medium impacts (and take land):
  ImpX:= (ImpactX-PosX)/Size;
  ImpY:= (ImpactY-PosY)/Size;
  if ((ImpX*cos(Angle) + ImpY*sin(Angle)) < -0.6) then begin
    Coef:= Power(max(0, -sin(Angle)), 4);
    Damage:= Damage / (1+Coef);
    Absorb:= Coef * 10;
  end else begin
    Absorb:= 5;
  end;
  Damage:= Max(0, Damage - Absorb);
  // Take the damage from the health:
  Health:= Health - Damage;
  // Landed?
  if (not dead) and (Health>0) and (sin(Angle)<-0.95) and
    // Both lower edges must be landed, don't do malabarisms dude
    (not TTrack2D(Track).IsInsideTrack(Nodes[2].X, Nodes[2].Y + 2)) and
    (not TTrack2D(Track).IsInsideTrack(Nodes[3].X, Nodes[3].Y + 2)) then begin
    if (Energy<50) and (not TTrack(Track).Evolution) then begin
      // Switch off engine before re-fuelling:
      MainPower:= 0;
      RotaPower:= 0;
      Energy:= Energy + 1;
      // Give a sweet to the bird if it landed safely (score depends on health
      // so lossing 50% is not appealing) so it land ok and stay on my hands:
      State.AddScore( Health/100 );
    end else begin
      MainPower:= Max(20, MainPower); // Give it extra impulse to take off safely
    end;
  end;
end;


// *********************
// ** EVOLUTION STUFF **
// *********************

function TRocket.MakeADuplicate: TPlayer;
begin
  // Duplicate yourself:
  result:= TRocket.create(0, 0, 0);
  TRocket(result).CopyFrom(self);
end;

// *******************
// ** SET/GET STUFF **
// *******************

procedure TRocket.SetMainPower(v: double);
begin
  State.Param[fMainPower].Value:= v;
end;
function  TRocket.GetMainPower: double;
begin
  result:= State.Param[fMainPower].Value;
end;

procedure TRocket.SetRotaPower(v: double);
begin
  State.Param[fRotaPower].Value:= v;
end;
function  TRocket.GetRotaPower: double;
begin
  result:= State.Param[fRotaPower].Value;
end;


procedure TRocket.SetMainPush(v: double);
begin
  State.Param[fMainPush].Value:= v;
end;
function  TRocket.GetMainPush: double;
begin
  result:= State.Param[fMainPush].Value;
end;

procedure TRocket.SetRotaPush(v: double);
begin
  State.Param[fRotaPush].Value:= v;
end;
function  TRocket.GetRotaPush: double;
begin
  result:= State.Param[fRotaPush].Value;
end;

end.
