unit UState;

// A TPlayer will always has an "State": it's position, velocity, angle, etc.
// so a state will be a compound of "StateParams" (position.x is one)
// defining the actual "situation" or state of the TPlayer.
//
// StateParams can be of tree types:
//
//   1) ptFixed = Fixed params, it's value won't change with time, like Size,
//      Color, MaxPower, etc. Usually you set them on creation and from
//      there you just read them as you need.
//   2) ptComputed = The values change with time by the simulation itself,
//      like the car position or its velocity.
//   3) ptFree: Correspond to a "degree of freedom", it's value is decided
//      by the AI, like the trottle and the driving wheel position in a kart.
//      Think of them as the "channels" in a RC model, or how many directions
//      you can move the joystick in the RC controler.
//      A RC car has 2 of them (accelerator/turning) while and helicopter can
//      have 3 to 5 depending on models.
//
//      Example: A kart has a computed param called "Turn" that reflects the
//      actual angle of the front tires (or the position of the driving wheel).
//      The AI can't decide the "new position" of the driving wheel, what it
//      decides is the "Force" it will apply to the driving wheel to make it
//      change. This "Turning push" is the "free" param. that AI will fill with
//      the "intelligent" value, and the simultation will use it as a force to
//      change the turning wheel position. Think of it like this: When you drive
//      your car you push left or right with a given force, and push accelator
//      harder or softer, but you don't directly "place" the driving wheel or the
//      acceleator in a decided position.
//
//      Note: If a player has no "Free" params then it is a dumb player, like a
//      rock rolling down the hill, that you need to simulate but it won't
//      take any "decision".
//
// Each StateParam will has a "Precision" that must be interpeted on computed
// ones as a "Grid Size" that makes a point to snap to a imaginary grid.
// Precision = 0 means it is discarted on the MacroState conversion.
//
// Also, to avoid the need of storing the "old" and "initial" values in a copy
// of the "State", each param can keep a copy of its old value so you can get
// "deltas" and use them to scrore the changes as you move on into a future.
//
// The initial values are mainly used to reset the simulation to the starting
// point in the user interface, but also has a "real" use during the thinking
// proccess: Once you get 100 future end points for a player, you have to group
// them so all futures with the same initial decision are together. Only inside
// those groups of futures you can discard similar ending points. So the initial
// decision (values of "Free" params for the first frame in the thinking
// proccess) need to be "discretized" so you only have a few choices and safely
// stored in the State so you can retrieve in the decision-making phase,
// so the FreeParam.IniValue will be set on the first move and not changed
// during the rest of the future imagining proccess.


interface

uses Contnrs, Variants;

type TParamType = ( ptFixed, ptComputed, ptPositional, ptFree );

// Each component on the player's state is a TStateParam: PosX, Angle, VelX, etc.
//
// For each of them I will store 3 values: Value is the actual one, OldValue is
// the previous one (1 delta time ago), and IniValue is the value in the start.
type TStateParam = class(TObject)
  private
    // The name for this param: "PosX", "Heading", "VelX", etc.
    fName: string;
    // The value itself: angle, coordinate, etc.
    fValue: double;
    // The "old" value (in previous frame) so you can compare and get deltas:
    fOldValue: double;
    // The "initial" value is used in two cases:
    // 1) If you want to "reset" your player to the initial position, you need
    //    to revert all StateParam to theirs ini values.
    // 2) In the case of a param that is a "degree of freedom" (AI can decide
    //    to push it up or down, like the driving wheel angle in a RC Car) and
    //    you are computing a possible future, you need to keep this initial
    //    "decision" so you can ponder later if it was good or bad.
    // Internally, we use integer and round(Value*1000) so comparations are ok.
    fIniValue: integer;
    // Some times you want to memorize a value for later use, like M+ in a
    // calculator. For instance, to bounce back to track it is nice to have
    // the "last in-track known position" to move in that direction the player
    // to avoid keeping outside track after a bounce calculation.
    fMemoValue: double;
    // There are 3 kind of components as stated above: Params, Computed by the
    // simulation and Free (degree of freedom):
    fParamType: TParamType;
    // Boundaries of allowed values for that param (both zero for unbounded):
    fMaxValue, fMinValue: double;
    // You can take random values with a flat or a Gaussian distribution:
    fUseGaussian: boolean;
    // Precision when converted into a micro-state (use 0 to discard it):
    // On free params, it is the "small option" or standard deviation for random.
    fPrecision: double;
    // On free params "sensitivity" modifiy "small value" so you can adjust it
    // low for precission maniouvres:
    fSensitivity: double;
    // Integer version of the RoundValue so comparing is secure and fast.
    fRoundValue: integer;
    procedure MicroToMacro;
    procedure SetValue(v: double);
    procedure SetIniValue;
    function  GetIniValue: double;
    procedure SetName(v: string);
    function  GetSmallOption: double;
    function  GetBigOption: double;
  public
    // Constructor:
    constructor create(vName: string; vType: TParamType; vValue, vMax, vMin, vPrec: double); overload;
    constructor createFixed(vName: string; vValue, vMax, vMin: double);
    constructor createComputed(vName: string; vValue, vMax, vMin: double);
    constructor createPositional(vName: string; vValue, vPrec: double);
    constructor createFree(vName: string; vMax: double);
    // Before copying form another state I need to create empty StateParams:
    constructor create; overload;
    // Copy from another param:
    procedure CopyFrom(Other: TStateParam);
    // Memorize actual value for later use:
    procedure MemorizeValue;
    // Is it involved in decision making or macro converting?
    function IsFreeParam: boolean;
    function IsPositionalParam: boolean;
    // Take a random decision (initail one is different form others):
    procedure TakeFirstDecision;
    procedure TakeRandomDecision;
    // Comparing initial decision is better with integers:
    function  IniValueAsInteger: double;
    // Public porperties:
    property Value:      double  read fValue    write SetValue;
    property OldValue:   double  read fOldValue write fOldValue;
    property IniValue:   double  read GetIniValue;
    property MemoValue:  double  read fMemoValue write fMemoValue;
    property RoundValue: integer read fRoundValue;
    property ParamType: TParamType read FParamType;
    property UseGaussian: boolean read fUseGaussian write fUseGaussian;
    property Name: string read FName write SetName;
    property Precision: double read fPrecision write fPrecision;
    property SmallOption: double read GetSmallOption;
    property BigOption: double read GetBigOption;
    property Sensitivity: double read fSensitivity write fSensitivity;
    property MaxValue:  double read fMaxValue  write fMaxValue;
    property MinValue:  double read fMinValue  write fMinValue;
  end;


// All values of a player's state (the StateParams) form it's TState:
type TState = class(TObject)
  private
    // The TPlayer this state belongs to:
    fOwner: TObject;
    // General list of all its params of any type:
    fParams: TObjectList;
    // Sublist with only free params (to speed up)
    fFreeParams: TObjectList;
    // Sublist with Positional (micro-to-macro) params:
    fPositionalParams: TObjectList;
    // A State can be a future and belong to an "option",  so a pointer will help
    fBelongsTo: TObject;
    // A future can be discarted if it is repeated:
    fFutureDiscarted: boolean;
    // Use gaussian distribution on random decision for all params:
    fUseGaussian: boolean;
    //
    // When a state is used as a future, you have 3 scorings numbers that will
    // form the "final" score used to decide:
    //
    //  1) ScorePos : How good -positive- the future is (raced distance)
    //  2) ScoreRed : Reduction coef., usually is 1 but a collision can reduce it.
    //  3) ScoreNeg : How bad -negative- the future is (health losed).
    fScorePos, fScoreRed, fScoreNeg: extended;
    // A final of future sate can be reached by several futures (not in use):
    fHitCount: integer;
    // How long have this future traveled (milliseconds)?
    fTimePassed: int64;
    // If you died in this future, how long did you survived?
    fTimeSurvived: int64;
    // Get State param by index or name:
    function GetParam(i: integer): TStateParam;
    function GetFreeParam(i: integer): TStateParam;
    function GetPositionalParam(i: integer): TStateParam;
    function GetCanDecide: boolean;
    function GetParamByName(v: string): TStateParam;
    function GetScore: extended;
    procedure SetUseGaussian(v: boolean);
  protected
    property ScorePos: extended read fScorePos write fScorePos;
    property ScoreRed: extended read fScoreRed write fScoreRed;
    property ScoreNeg: extended read fScoreNeg write fScoreNeg;
  public
    // Create and destroy it:
    constructor create(MyOwner: TObject);
    destructor  destroy; override;
    // Add another StateParam:
    procedure AddParam(v: TStateParam);
    // Copy values form another TState:
    procedure CopyFrom(Other: TState);
    // Number of params of this state:
    function ParamCount: integer;
    function FreeParamCount: integer;
    function PositionalParamCount: integer;
    // Debug only:
    function  GetDecisionVector: string;
    // Set the OldValues to Value.
    procedure SetOldState;
    // Go back to the previous or ini state:
    procedure BackToOldState;
    procedure BackToIniState;
    // Take actual state as your "initial" -and thereof your "old"- state:
    procedure SetIniState;
    // Take a random decision (initail one is different form others):
    procedure TakeFirstDecision;
    procedure TakeRandomDecision;
    // Round final state of a future to a macro estate (only affect Positional params):
    procedure MicroToMacro;
    // Future comparation stuff:
    property FutureDiscarted: boolean read fFutureDiscarted write fFutureDiscarted;
    function HaveSameInitialDecision(Other: TState): boolean;
    function HaveSameFinalState(Other: TState): boolean;
    // Adding a goal scoring to my score:
    procedure AddScore(points: extended; reduction: extended = 1);
    // Reset scoring to a fixed value:
    procedure SetScore(v: extended);
    // Easy access to a param by its index:
    property Owner: TObject read fOwner;
    property BelongsTo: TObject read fBelongsTo write fBelongsTo;
    property Param[i: integer]: TStateParam read GetParam;
    property FreeParam[i: integer]: TStateParam read GetFreeParam;
    property PositionalParam[i: integer]: TStateParam read GetPositionalParam;
    // Easy access to a param by its name:
    function ParamIndexByName(v: string): integer;
    property ParamByName[n: string]: TStateParam read GetParamByName;
    // How "far" have the player moved is the "score" of the future state:
    property Score: extended read GetScore; // write SetScore;
    property HitCount: integer read fHitCount write fHitCount;
    property TimePassed: int64 read fTimePassed write fTimePassed;
    property TimeSurvived: int64 read fTimeSurvived write fTimeSurvived;
    property UseGaussian: boolean read fUseGaussian write SetUseGaussian;
    procedure ResetScore;
    // Does this player have any free param to decide on?
    property CanDecide: boolean read GetCanDecide;
  end;


implementation

uses UPlayer, Math, SysUtils, Classes;


// *****************
// ** TStateParam **
// *****************

// **************************
// **     TStateParam      **
// ** ADMINISTRATIVE STUFF **
// **************************

// Complete create with all values set on origin:
constructor TStateParam.create(vName: string; vType: TParamType; vValue, vMax, vMin, vPrec: double);
begin
  inherited create;
  // Initialize:
  Name:=       vName;
  fParamType:= vType;
  fMaxValue:=  vMax;
  fMinValue:=  vMin;
  Value:=      vValue;
  OldValue:=   vValue;
  MemoValue:=  vValue;
  SetIniValue;
  fUseGaussian:= true;
  fPrecision:= vPrec;
  fSensitivity:= 0.5; // is used x2 so 0.5 = neutral
end;

constructor TStateParam.createFixed(vName: string; vValue, vMax, vMin: double);
begin
  create(vName, ptFixed, vValue, vMax, vMin, 0);
end;
constructor TStateParam.createComputed(vName: string; vValue, vMax, vMin: double);
begin
  create(vName, ptComputed, vValue, vMax, vMin, 0);
end;
constructor TStateParam.createPositional(vName: string; vValue, vPrec: double);
begin
  create(vName, ptPositional, vValue, 0, 0, vPrec);
end;
constructor TStateParam.createFree(vName: string; vMax: double);
begin
  create(vName, ptFree, 0, vMax, -vMax, 0);
end;

// Fake create to get an empty StateParam that needs to be "copyied from"
constructor TStateParam.create;
begin
  create('', ptFixed, 0.0, 0.0, 0.0, 0.0);
end;

procedure TStateParam.SetValue(v: double);
begin
  // Boundaries are active when they "make sense" (two zeros doesn't make sense)
  if (fMaxValue > fMinValue) then
    // Avoid out of boundaries values:
    fValue:= Min(fMaxValue, Max(fMinValue, v))
  else
    fValue:= v;
end;

// Use actual value as the IniValue, rounding it first (ini values are rounded
// for better comparation) but keep 4 decimals:
procedure TStateParam.SetIniValue;
begin
  fIniValue:= round(1000*Value);
end;

function TStateParam.GetIniValue: double;
begin
  result:= fIniValue / 1000;
end;
function TStateParam.IniValueAsInteger: double;
begin
  result:= fIniValue;
end;

// Names are stored in upper case and trimmed:
procedure TStateParam.SetName(v: string);
begin
  fName:= UpperCase(Trim(v));
end;

// Copy from another param:
procedure TStateParam.CopyFrom(Other: TStateParam);
begin
  fName:=       Other.Name;
  fParamType:=  Other.FParamType;
  fMaxValue:=   Other.MaxValue;
  fMinValue:=   Other.MinValue;
  fPrecision:=  Other.Precision;
  fSensitivity:=Other.Sensitivity;
  fValue:=      Other.fValue;
  fOldValue:=   Other.fOldValue;
  fMemoValue:=  Other.MemoValue;
  fIniValue:=   Other.fIniValue;
  fUseGaussian:=Other.UseGaussian;
end;

// Memorize actual cvalue for later use:
procedure TStateParam.MemorizeValue;
begin
  fMemoValue:= Value;
end;

// Is it involved in decision making?
function TStateParam.IsFreeParam: boolean;
begin
  result:= (ParamType = ptFree);
end;
// Is it involved in macro state (Positional)?
function TStateParam.IsPositionalParam: boolean;
begin
  result:= (ParamType = ptPositional);
end;

// ********************
// **  TStateParam   **
// ** THINKING STUFF **
// ********************

// Take the initial random decision:
// First decision must be discretized and stored for later analysis.
procedure TStateParam.TakeFirstDecision;
begin
  // First decisions must be "discretized", so i choose between 4 values:
  case floor(4*random) of
    0: Value:= +BigOption;
    1: Value:= -BigOption;
    2: Value:= +SmallOption;
    3: Value:= -SmallOption;
  end;
  // Store as IniValue (the initial decision -or "option- of the future):
  SetIniValue;
end;

// Take another random decision:
procedure TStateParam.TakeRandomDecision;
begin
  if UseGaussian then begin
    // Using a gaussian bell distribution:
    Value:= RandG(0, SmallOption);
  end else begin
    // Using a flat distribution:
    Value:= (2*random-1) * SmallOption;
  end;
end;

function TStateParam.GetSmallOption: double;
begin
  result:= (MaxValue/3) * (2*fSensitivity);
end;

function TStateParam.GetBigOption: double;
begin
  result:= MaxValue * (2*fSensitivity);
end;

// Convert a MicroState value to MacroState:
// Value 2.45, precision 0.5, then RoundValue is 5 as int, and Value is now 2.5
procedure TStateParam.MicroToMacro;
begin
  if (Precision = 0) then
    raise Exception.create('Precision zero param "'+Name+'" can not be converted to macro.');
  fRoundValue:= round(Value/Precision);
  Value:= Precision * fRoundValue;
end;


// ************
// ** TState **
// ************

constructor TState.create(MyOwner: TObject);
begin
  // Is owner a TPlayer?
  if not (MyOwner is TPlayer) then
    raise Exception.create('State''s owner is not a TPlayer.');
  // Create the base object:
  inherited create;
  // Remeber the owner:
  fOwner:= MyOwner;
  // Still not in any group:
  fBelongsTo:= nil;
  fFutureDiscarted:= false;
  // Use falt distribution for random:
  fUseGaussian:= false;
  // Initialize:
  fParams:= TObjectList.create;
  fParams.OwnsObjects:= false; // I create it, I kill it
  // Sub-list with only free params:
  fFreeParams:= TObjectList.create;
  fFreeParams.OwnsObjects:= false;
  // Sub-list with only Positional params:
  fPositionalParams:= TObjectList.create;
  fPositionalParams.OwnsObjects:= false;
  // Scoring coefs for futures:
  ResetScore;
end;

destructor TState.destroy;
begin
  // Free all the StateParams:
  while fParams.Count > 0 do begin
    TStateParam(fParams.Items[0]).Free;
    fParams.Delete(0);
  end;
  fParams.Free;
  fFreeParams.Free;
  fPositionalParams.Free;
end;

procedure TState.AddParam(v: TStateParam);
begin
  fParams.Add(v);
  // Is it a free param?
  if v.IsFreeParam then
    fFreeParams.Add(v);
  // Is it Positional = related to micro to macro?
  if v.IsPositionalParam then
    fPositionalParams.Add(v);
end;

function TState.GetParam(i: integer): TStateParam;
begin
  if ((i < 0) or (i> ParamCount-1)) then
    raise Exception.create('TState: Param index '+IntToStr(i)+' out of boundaries.');
  result:= TStateParam(fParams[i]);
end;

function TState.GetFreeParam(i: integer): TStateParam;
begin
  if ((i < 0) or (i > FreeParamCount-1)) then
    raise Exception.create('TState: FreeParam index '+IntToStr(i)+' out of boundaries.');
  result:= TStateParam(fFreeParams[i]);
end;

function TState.GetPositionalParam(i: integer): TStateParam;
begin
  if ((i<0) or (i>fPositionalParams.Count-1)) then
    raise Exception.create('TState: Positional Param index '+IntToStr(i)+' out of boundaries.');
  result:= TStateParam(fPositionalParams[i]);
end;

function TState.GetCanDecide: boolean;
begin
  result:= (fFreeParams.Count>0);
end;

// Witch index correspond to a param with this name?
function TState.ParamIndexByName(v: string): integer;
var i: integer;
begin
  v:= UpperCase(Trim(v));
  for i:= 0 to ParamCount-1 do begin
    if (v = Param[i].Name) then begin
      result:= i;
      exit;
    end;
  end;
  raise Exception.create('TState: Param with name "'+v+'" not found.');
end;

// This way to access params is easy but slow, keep in mind:
function TState.GetParamByName(v: string): TStateParam;
begin
  result:= Param[ParamIndexByName(v)];
end;


// Copy values form another TState:
procedure TState.CopyFrom(Other: TState);
var i: integer;
begin
  // Both states need to have the same number of params:
  while (ParamCount > Other.ParamCount) do
    fParams.Delete(fParams.Count-1);
  while (ParamCount < Other.ParamCount) do
    AddParam(TStateParam.create());
  // Clear free params' list:
  fFreeParams.Clear;
  fPositionalParams.Clear;
  // Copy all param's info:
  for i:= 0 to ParamCount-1 do begin
    Param[i].CopyFrom(Other.Param[i]);
    if Param[i].IsFreeParam then
      fFreeParams.Add(Param[i]);
    if Param[i].IsPositionalParam then
      fPositionalParams.Add(Param[i]);
  end;
  // Other info:
  UseGaussian:= Other.UseGaussian;
  // Copy its scoring:
  ScorePos:= Other.ScorePos;
  ScoreNeg:= Other.ScoreNeg;
  ScoreRed:= Other.ScoreRed;
  HitCount:= Other.HitCount;
end;

// Reset the OldValue in all components of this state.
// Why? After a big state change (you manually move the thing) keeping the old
// state is just dangerous, as it is no longer true.
procedure TState.SetOldState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do
    Param[i].OldValue:= Param[i].Value;
end;

// Number of params of this state:
function TState.ParamCount: integer;
begin
  result:= fParams.Count;
end;

// Number of FREE params of this state:
function TState.FreeParamCount: integer;
begin
  result:= fFreeParams.Count;
end;

// Number of Positional (Macro related) params of this state:
function TState.PositionalParamCount: integer;
begin
  result:= fPositionalParams.Count;
end;

// Go back to the previous state:
// Why? You need to fine-adjust actual position without disturbing the old state,
// for instance, 2 players crash into each other and you want to "walk back" in
// time a little to the exact moment where contact occurred.
procedure TState.BackToOldState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do
    Param[i].Value:= Param[i].OldValue;
end;

// Go back to the initial state (reset simulation):
procedure TState.BackToIniState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do
    Param[i].Value:= Param[i].IniValue;
end;

// Set actual state as the "initial" state of the player.
procedure TState.SetIniState;
var i: integer;
begin
  for i:= 0 to ParamCount-1 do begin
    Param[i].SetIniValue;
    Param[i].OldValue:= Param[i].Value;
  end;
end;

// Convert this state from a micro-state to a macro-state
procedure TState.MicroToMacro;
var i: integer;
begin
  // Convert each related param from micro to macro state by rounding it:
  for i:= 0 to PositionalParamCount-1 do
    PositionalParam[i].MicroToMacro;
end;

// Take first random decision of a future (only set one of the freeparams):
procedure TState.TakeFirstDecision;
var i: integer;
begin
  // No free params? Then it can't decide anything, don't waste time:
  if CanDecide then begin
    // I only decide on ONE 1 free param:
    i:= floor(random * FreeParamCount);
    // Ask it to take a decision:
    FreeParam[i].TakeFirstDecision;
  end;
end;

// Take a random decision on all free params:
procedure TState.TakeRandomDecision;
var i: integer;
begin
  if CanDecide then
    for i:= 0 to FreeParamCount-1 do
      FreeParam[i].TakeRandomDecision;
end;

procedure TState.SetUseGaussian(v: boolean);
var i: integer;
begin
  fUseGaussian:= v;
  for i:= 0 to FreeParamCount-1 do
    FreeParam[i].UseGaussian:= v;
end;

// Get the final score of the future:
function TState.GetScore: extended;
begin
  // AI_Level changes the way Scores (positive, negative and reduction) are mixed:
  case TPlayer(Owner).AI_Level of
    1: // Level 1: Only "good" ends (score>0.5) count as 1 (using S = N):
      if (TPlayer(Owner).Dead) then result:= 0 else result:= 1;
    2: // Level 2: All futures count as 1 (using S = N)
      result:= 1;
    3: // Level 3: All futures count as 1 (using S = k*Ln(N))
      result:= 1;
    4: // Level 4: Use "Score" as it comes (using S = Sum(raced distance))
       // Reductions can be used for first!
      result:= ScorePos * ScoreRed;
    5,6: // Level 5 and up: Use "Score^2" (using S = Sum(Score^2)).
         // Reductions must also be squared. Negatives are not allowed.
      result:= Sqr(ScorePos * ScoreRed);
    7,8: // Level 7: Worst option will score 0, so negative scoring is allowed:
         // Negative scores doesn't apply reductions, seems logical...
      result:= Sqr(ScorePos * ScoreRed) - Sqr(ScoreNeg);
    else // There should be no "else" case.
      raise Exception.Create('Using an AI level not defined in UState.GetScore!');
  end;
  {
  // Mixing positive goals with reductive goals is only feasible from level 4,
  // lower levels use 0/1 so reduction is not possible:
  if TPlayer(Owner).AI_Level >= 4 then
    result:= result * ScoreRed;
  // Add NegScore to the mix AFTER mixing positive with reductions (neg doesn't
  // reduct) and only for AI level 7 or higher. Only level 7 use
  // score=score-minimum so negatives convert into zero or positives.
  if TPlayer(Owner).AI_Level >= 7 then
    result:= result + ScoreNeg;
  }
end;

// Some times you need to set a new score from scratch:
procedure TState.SetScore(v: extended);
begin
  ScorePos:= Max(0, v);
  ScoreNeg:= Min(0, v);
  ScoreRed:= 1;
end;

procedure TState.ResetScore;
begin
  ScorePos:= 0;
  ScoreRed:= 1;
  ScoreNeg:= 0;
  HitCount:= 0;
  TimePassed:= 0;
  TimeSurvived:= -1; // -1 means "Hey I did die!"
end;

procedure TState.AddScore(points: extended; reduction: extended = 1);
begin
  // Positive scores goes separated from negative ones:
  ScorePos:= ScorePos + Max(0, points);
  ScoreNeg:= ScoreNeg - Min(0, points);
  // Also accumulate (product) those multipliers:
  ScoreRed:= ScoreRed * reduction;
end;

function TState.GetDecisionVector: string;
var i: integer;
begin
  result:= '';
  for i:= 0 to FreeParamCount-1 do begin
    if (result<>'') then result:= result + ', ';
    result:= result + IntToStr(round(FreeParam[i].IniValue));
  end;
  result:= '(' +result+') #'+IntToStr(HitCount);
end;

// ********************
// **     TState     **
// ** THINKING STUFF **
// ********************

// Two futures with same initial decision belongs to the same "option" so they
// are two possible futures endings for a single initial decision:
function TState.HaveSameInitialDecision(Other: TState): boolean;
var i: integer;
begin
  // Each free param must have identical initial value:
  for i:= 0 to FreeParamCount-1 do begin
    if (FreeParam[i].IniValueAsInteger <> Other.FreeParam[i].IniValueAsInteger) then begin
      // Ops, they don't share the initial decision for at least a free param!
      result:= false;
      exit;
    end;
  end;
  // No ofending free param was found, they share a starting point!
  result:= true;
end;

// Two futures that end somehow "near" represent the same micro state.
function TState.HaveSameFinalState(Other: TState): boolean;
var i: integer;
begin
  // Each Positional param (macro related param) must have identical value:
  for i:= 0 to PositionalParamCount-1 do begin
    if (PositionalParam[i].RoundValue <> Other.PositionalParam[i].RoundValue) then begin
      // They don't end in the same MacroState!
      result:= false;
      exit;
    end;
  end;
  // No difference was found, they share the final MacroState!
  result:= true;
end;

end.
