unit UTrack;

// TRACK is the place where the simulation runs, the chessboard, it knows
// with teams and individual players exists, check for collisions and acts
// as a director on the thinking proccess by syncing all players' future
// creation.

interface

uses ExtCtrls, Contnrs, Forms, StdCtrls,
     UTeam;

type
  TTrack = class(TObject)
  private
    fTeams: TObjectList;
    fSelectedTeam: TTeam;
    // How many milliseconds in the future look ahedad when thinking?
    fFutureMSeconds: integer;
    fSimMSeconds: integer;
    // How many futures will I compute before deciding?
    fMinFutures: integer;
    // Detect collisions player-player?
    fPlayersCollide: boolean;
    // Activate evolution?
    fEvolution: boolean;
    // Property getters
    function GetTeam(i: integer): TTeam;
    function GetTeamCount: integer;
    function GetPlayerCount: integer;
    function GetSimSeconds: double;
    procedure SetEvolution(v: boolean);
  protected
    fPlayers: TObjectList;
    procedure DetectCollisions; virtual; abstract;
    procedure ClearTrack; virtual; abstract;
  public
    constructor Create;
    destructor  Destroy; override;
    // Make all the players to think and move for one tick:
    procedure TakeAStep(CPU_time: integer);
    procedure DelTeam(v: TTeam; DoFree: boolean = false);
    // Draw track and all players on theirs positions:
    // Note: "Clear" image before drawing is an option here. During simulation,
    // it is cleared before starting calcs, during the calcs. debug info is added
    // and finally, when all is decided, players are added calling draw(false):
    procedure Draw(clearing: boolean);
    procedure DrawOptions;
    procedure AddTeam(v: TTeam);
    procedure RefreshPlayerList;
    property  Team[i: integer]: TTeam read GetTeam;
    property  TeamCount: integer read GetTeamCount;
    property  SelectedTeam: TTeam read fSelectedTeam write fSelectedTeam;
    property  PlayerCount: integer read GetPlayerCount;
    property  MinFutures: integer read fMinFutures write fMinFutures;
    property  FutureMSeconds: integer read fFutureMSeconds write fFutureMSeconds;
    property  SimMSeconds: integer read fSimMSeconds write fSimMSeconds;
    property  SimSeconds: double read GetSimSeconds;
    property  PlayersCollide: boolean read fPlayersCollide write fPlayersCollide;
    property  Evolution: boolean read fEvolution write SetEvolution;
  end;


implementation

uses DateUtils, SysUtils, Graphics, Types, Math,
     UUtils, UPlayer, UGoal;


// **************************
// ** ADMINISTRATIVE STUFF **
// **************************

constructor TTrack.Create;
begin
  inherited;
  // Teams on track:
  fTeams:= TObjectList.create;
  fTeams.OwnsObjects:= false;
  fSelectedTeam:= nil;
  // Players on the track:
  // Note: It is redundant with team list, but checking collisions needs it.
  fPlayers:= TObjectList.create;
  fPlayers.OwnsObjects:= false;
  // Futures to compute and for how many seconds:
  fMinFutures:= 100;
  fFutureMSeconds:= 10*500; // 5s default
  fSimMSeconds:= 100; // 0.1s default
  // Activate evolution?
  Evolution:= false;
end;

destructor TTrack.destroy;
begin
  FTeams.Free;
  FPlayers.free;
  inherited;
end;


// ********************
// ** MANAGING TEAMS **
// ********************

function TTrack.GetTeam(i: integer): TTeam;
begin
  result:= TTeam(FTeams[i]);
end;

function TTrack.GetTeamCount: integer;
begin
  result:= FTeams.Count;
end;

// Add a team to the track:
procedure TTrack.AddTeam(v: TTeam);
begin
  FTeams.Add(v);
  v.Track:= Self;
  SelectedTeam:= v;
end;

// Delete a team from the track:
procedure TTrack.DelTeam(v: TTeam; DoFree: boolean = false);
var p: integer;
begin
  p:= FTeams.IndexOf(v);
  if (p <> -1) then begin
    if DoFree then
      Team[p].free;
    FTeams.Delete(p);
  end;
end;


// ******************************
// ** LIST OF PLAYERS ON TRACK **
// ******************************

function TTrack.GetPlayerCount: integer;
begin
  result:= fPlayers.Count;
end;

// Add/Del players using the track instead of the team, as the track needs the
// list of all players on it in order to detect collisions of any 2 players.
procedure TTrack.RefreshPlayerList;
var t, p: integer;
begin
  fPlayers.Clear;
  for t:= 0 to TeamCount-1 do begin
    for p:= 0 to Team[t].PlayerCount-1 do begin
      fPlayers.Add(Team[t].Player[p]);
    end;
  end;
end;

procedure TTrack.SetEvolution(v: boolean);
var t, p: integer;
begin
  fEvolution:= v;
  if Evolution then begin
    for p:= 0 to PlayerCount-1 do begin
      TPlayer(fPlayers[p]).ColorUsingGoals;
      TPlayer(fPlayers[p]).DebugGoals:= true;
    end;
  end;
end;

// ********************
// ** THINKING STUFF **
// ********************

// TODO : AutoAdjust MinFutures if a player not having too many options, so
//        before it crash, it is given more time to find a solution.

// Make all players to think, decide, and move (do a complete frame):
// Note: All times in milliseconds.
procedure TTrack.TakeAStep(CPU_time: integer);
var
  p, t, ticks: integer;
  FoundFutures: integer;
  EndTime: LongWord;
begin
  // Number of ticks needed to complete a future:
  ticks:= Max(2, round(FutureMSeconds/SimMSeconds));
  // Reset the list of futures found:
  for p:= 0 to TeamCount-1 do
    Team[p].ResetFutures;
  // Try to allow 10% of the CPU time to analizing futures and move:
  EndTime:= MillisecondOfTheDay(Now) + round(0.9*CPU_time);
  // Clear the track image:
  ClearTrack;
  // Compute futures until time expires and you have the minimun amount:
  FoundFutures:= 0;
  repeat
    // All players enter "imagining mode" (so a new future is created) and take
    // the intial random decision (it must be stored for later use):
    for p:= 0 to TeamCount-1 do
      Team[p].Imagining:= true;
    // Take "sim_time" steps until you get to FutureSeconds
    for t:= 1 to ticks do begin
      // Ask each player to advance it's future a delta-time:
      for p:= 0 to TeamCount-1 do
        if (t=1) then
          Team[p].TakeFirstStep
        else
          Team[p].TakeRandomStep;
      // Detect and react to collisions:
      DetectCollisions;
      // Score state changes on all teams and all players:
      for p:= 0 to TeamCount-1 do begin
        Team[p].ScoreGoals(gfStep);
        // Draw goals that need drawing and don't draw on eachplayer:
        Team[p].DrawGoals;
      end;
    end;
    // End imagining this future:
    for p:= 0 to TeamCount-1 do
      Team[p].Imagining:= false;
    // Another future computed!
    inc(FoundFutures);
    // Make GUI responsible:
    //Application.ProcessMessages;
  until (MillisecondOfTheDay(Now) > EndTime) and (FoundFutures > MinFutures);
  //
  // Score state changes on all teams and all players:
  for p:= 0 to TeamCount-1 do begin
    Team[p].ScoreGoals(gfStep);
    // Draw goals that need drawing and don't draw on eachplayer:
    Team[p].DrawGoals;
  end;
  // Now players are back to reality.
  // Analize the futures found and decide in a "intelligent" way:
  for p:= 0 to TeamCount-1 do
    Team[p].TakeIntelligentStep;
  // Detect "real" collisions between any two players and tell them to react:
  DetectCollisions;
  // Evolution?
  if Evolution then begin
    for p:= 0 to TeamCount-1 do
      Team[p].Duplicate;
  end;

  // Dead players with size<1 dissapear from the game:
  for p:= 0 to TeamCount-1 do
    Team[p].DeleteNosized;

  // Screen must be updated if debug info was draw:
  Draw(false);
end;

function TTrack.GetSimSeconds: double;
begin
  result:= SimMSeconds/1000;
end;


// Draw all players over the track:
procedure TTrack.Draw(clearing: boolean);
var
  t: integer;
begin
  // TODO: Use triple buffering, one for original track picture, another for
  // adding there trails (dots) and the 3rd to add the players on top.
  // This final 3rd one would also be double-buffered to avoid flickering.
  if clearing then
    ClearTrack;
  // Debugging ending points?
  DrawOptions;
  // Draw players on top of all:
  for t:= 0 to TeamCount-1 do
    Team[t].Draw;
  // Refresh screen:
  Application.ProcessMessages;
end;

procedure TTrack.DrawOptions;
var t: integer;
begin
  for t:= 0 to TeamCount-1 do
    Team[t].DrawOptions;
end;

end.
