unit UTrack2D;

interface

uses ExtCtrls, SysUtils,
     UTrack, UPlayer2D;

type
  TTrack2D = class(TTrack)
  private
    TrackScan: array of array of byte; //0=track, 1=out, 2=red recharge
    // General friction (avoid running too fast):
    FFriction: double;
    // Side friction (avoid sliding):
    FSideFriction: double;
    // Pointer to the track image so i can refresh it:
    FImage: TImage;
    // Image dimensions:
    FImageMaxX, FImageMaxY: integer;
    procedure ScanTrack(Im: TImage);
    function  GetPlayer2D(i: integer): TPlayer2D;
  protected
    procedure DetectCollisions; override;
  public
    CopyImage: TImage;
    constructor Create;
    function    MouseIn: boolean;
    function    MouseX: integer;
    function    MouseY: integer;
    function    IsInsideImage(x, y: double): boolean;
    function    IsInsideTrack(x, y: double): boolean;
    procedure   ClearTrack; override;
    property    Player2D[i: integer]: TPlayer2D read GetPlayer2D;
    property    Img: TImage read FImage write ScanTrack;
    property    Friction: double read FFriction write FFriction;
    property    SideFriction: double read FSideFriction write FSideFriction;
  end;

implementation

uses Graphics, Forms, Math, Controls, 
     UTeam, UUTils, UKart, URocket;


// ***************************
// ** ADMNINISTRATIVE STUFF **
// ***************************

constructor TTrack2D.Create;
begin
  inherited;
  FImage:= Nil;
  FFriction:= 0.015;
  FSideFriction:= 0.20;
end;

function TTrack2D.GetPlayer2D(i: integer): TPlayer2D;
begin
  if fPlayers[i] is TPlayer2D then
    result:= TPlayer2D(FPlayers[i])
  else
    raise Exception.create('Track2D contains a player that is not a Player2D.');
end;


// *******************
// ** DRAWING STUFF **
// *******************

procedure TTrack2D.ScanTrack(Im: TImage);
var
  i, j: integer;
begin
  // Remember it for later use:
  FImage:= Im;
  // Store dimensiones:
  FImageMaxX:= Im.Picture.Width;
  FImageMaxY:= Im.Picture.Height;
  // Scan picture for black pixles:
  SetLength(TrackScan, Im.Picture.Width);
  for i:= Low(TrackScan) to High(TrackScan) do begin
    SetLength(TrackScan[i], Im.Picture.Height);
    for j:= Low(TrackScan[i]) to High(TrackScan[i]) do begin
      TrackScan[i, j]:= 0;
      if (Im.Picture.Bitmap.Canvas.Pixels[i, j] = clBlack) then
        TrackScan[i, j]:= 1; // Outside track
      if (Im.Picture.Bitmap.Canvas.Pixels[i, j] = clRed) then
        TrackScan[i, j]:= 2; // Battery charging zone
    end;
  end;
  // I need a duplicate of the track so I won't paint over the original one.
  CopyImage:= TImage.Create(nil);
  CopyImage.Visible:= false;
  CopyImage.Width:=   Im.Width;
  CopyImage.Height:=  Im.Height;
  CopyImage.Left:=    Im.Left;
  CopyImage.Top:=     Im.Top;
  CopyImage.Picture.Assign(Im.Picture);
end;

procedure TTrack2D.ClearTrack;
begin
  if Assigned(FImage) then
    FImage.Picture.Assign(CopyImage.Picture);
end;

function TTrack2D.MouseIn: boolean;
var x, y: integer;
begin
  x:= MouseX;
  y:= MouseY;
  result:= (x>0) and (x<Img.Width) and (y>0) and (y<Img.Height);
end;

function TTrack2D.MouseX: integer;
begin
  result:= Img.ScreenToClient(Mouse.CursorPos).X;
end;

function TTrack2D.MouseY: integer;
begin
  result:= Img.ScreenToClient(Mouse.CursorPos).Y;
end;


// **********************
// ** SIMULATION STUFF **
// **********************

// Detect collision between players or player-fence and make them to react:
procedure TTrack2D.DetectCollisions;
var
  i, j, k: integer;
  FromItoJX, FromItoJY, Dist: double;
  DifVelX, DifVelY: double;
  hitIJ, hitJI, EnergyAdjusted: boolean;
  Bounce, Suck: double;
  AliveOne, DeadOne: integer;
  IsKart, IsRocket: integer;
begin
  // Player-player collisions are checked?:
  if PlayersCollide then begin
    for i:= 0 to PlayerCount-2 do begin
      for j:= (i+1) to PlayerCount-1 do begin
        // First check if a collision is even possible:
        if (Player2D[i].MinX <= Player2D[j].MaxX) and
           (Player2D[i].MaxX >= Player2D[j].MinX) and
           (Player2D[i].MinY <= Player2D[j].MaxY) and
           (Player2D[i].MaxY >= Player2D[j].MinY) then begin
          // A node of Pi is inside Pj's body?
          hitIJ:= Player2D[i].CheckCollision(Player2D[j]);
          if hitIJ then
            hitJI:= false
          else
            hitJI:= Player2D[j].CheckCollision(Player2D[i]);
          if hitIJ or hitJI then begin
            // Ops, players i and j did collide!
            // First thing I need is the vector from Player(i) to Player(j):
            FromItoJX:= Player2D[j].PosX - Player2D[i].PosX;
            FromItoJY:= Player2D[j].PosY - Player2D[i].PosY;
            // Distance from players:
            if (FromItoJX=0) and (FromItoJY=0) then
              Dist:= 1
            else
              Dist:= Sqrt(Sqr(FromItoJX)+Sqr(FromItoJY));
            // Vector distance normalized:
            FromItoJX:= FromItoJX / Dist;
            FromItoJY:= FromItoJY / Dist;
            // Difference of velocities:
            DifVelX:= (Player2D[i].VelX - Player2D[j].VelX);
            DifVelY:= (Player2D[i].VelY - Player2D[j].VelY);
            // Escalar product DiffVel*Dist to sum/dec on each player:
            Bounce:= Max(0, +(DifVelX * FromItoJX) + (DifVelY * FromItoJY));
            // Add bouncing velocity to both players so they now diverge:
            Player2D[i].VelX:= Player2D[i].VelX - FromItoJX * 0.45 * Bounce;
            Player2D[i].VelY:= Player2D[i].VelY - FromItoJY * 0.45 * Bounce;
            Player2D[j].VelX:= Player2D[j].VelX + FromItoJX * 0.45 * Bounce;
            Player2D[j].VelY:= Player2D[j].VelY + FromItoJY * 0.45 * Bounce;
            // Walk back until not touching:
            for k:= 1 to 4 do begin
              Player2D[i].HandMove( -2*FromItoJX, -2*FromItoJY, 0, false);
              Player2D[j].HandMove( +2*FromItoJX, +2*FromItoJY, 0, false);
              if (hitIJ and not Player2D[i].CheckCollision(Player2D[j])) or
                 (hitJI and not Player2D[j].CheckCollision(Player2D[i])) then
                break;
            end;
            // Bounce energy is used to adjust both players' energy:
            EnergyAdjusted:= false;
            if Evolution then begin
              {
              // Touch a dead player to eat it!
              // Only one can be dead, not both:
              if (Player2D[i].Dead or Player2D[j].Dead) and not (Player2D[i].Dead and Player2D[j].Dead) then begin
                if Player2D[i].Dead then begin
                  DeadOne:=  i;
                  AliveOne:= j;
                end else begin
                  DeadOne:=  j;
                  AliveOne:= i;
                end;
                // Take its energy: bigger ones gives more energy.
                Player2D[AliveOne].Health:= Player2D[AliveOne].Health + Max(0, Player2D[DeadOne].Size-1)*10;
                // DeadOne disapears (only in real sim, not imagining):
                if not Player2D[DeadOne].Imagining then
                  Player2D[DeadOne].Size:= Player2D[DeadOne].Size - 1;
              end;
              }
              // Kart touch rocket: Kart takes some energy from rocket:
              if (not Player2D[i].Dead) and (not Player2D[j].Dead) then begin
                IsKart:=  -1;
                IsRocket:= -1;
                if (Player2D[i] is TKart) then IsKart:= i;
                if (Player2D[j] is TKart) then IsKart:= j;
                if (Player2D[i] is TRocket) then IsRocket:= i;
                if (Player2D[j] is TRocket) then IsRocket:= j;
                if (IsKart<>-1) and (IsRocket<>-1) then begin
                  // Karts doesn't damage when crashing on a rockeck:
                  Player2D[IsKart].Health:= Player2D[IsKart].Health + Bounce;
                  // Suck some energy from the rocket ONLY if it is alive, or if
                  // I am imagining a future (I don't know if it is dead):
                  if (not Player2D[IsRocket].Dead) or (Player2D[IsRocket].Imagining) then begin
                    // Energy to transfer from rocket to kart: 25%
                    Bounce:= Min(25, Player2D[IsRocket].Energy);
                    Player2D[IsKart].Energy:=   Player2D[IsKart].Energy   + Bounce * 0.9; // 90% transfer eficiencie
                    Player2D[IsRocket].Energy:= Player2D[IsRocket].Energy - Bounce;

                    // Killing instinct? (test only) 
                    Player2D[IsKart].State.AddScore( 500, 1.5 );

                  end;
                end;
              end;
            end;
            if not EnergyAdjusted then begin
              // Bouncing decrease health on both players:
              Player2D[i].Health:= Player2D[i].Health - Bounce;
              Player2D[j].Health:= Player2D[j].Health - Bounce;
            end;
          end;
        end;
      end;
    end;
  end;
  // Collisons player-track limits:
  for i:= 0 to PlayerCount-1 do begin
    Player2D[i].IsInsideTrack;
    Player2D[i].BounceWithFence;
  end;
end;

function TTrack2D.IsInsideImage(x, y: double): boolean;
var i, j: integer;
begin
  i:= round(x);
  j:= round(y);
  result:= (i>0) and (i<FImageMaxX) and (j>0) and (j<FImageMaxY);
end;

function TTrack2D.IsInsideTrack(x, y: double): boolean;
var i, j: integer;
begin
  i:= round(x);
  j:= round(y);
  if IsInsideImage(i, j) then
    result:= (TrackScan[i,j]<>1)
  else
    result:= false;
end;


end.
